"""!@package ep_bolfi.models.assess_effective_parameters
Evaluates a parameter set for, e.g., overpotential and capacity.
"""

import numpy as np
from scipy.optimize import root_scalar
from tabulate import tabulate
import pybamm
# Reset the PyBaMM colour scheme.
import matplotlib.pyplot as plt
plt.style.use("default")


class Effective_Parameters:
    """!@brief Calculates, stores and prints effective parameters. """

    def __init__(self, parameters, working_electrode='both'):
        """!@brief Preprocesses the model parameters.
        @param parameters
            A dictionary of parameter values with the parameter names as
            keys. For these names, please refer to
            ep_bolfi.models.standard_parameters.
        @param working_electrode
            When set to either 'negative' or 'positive', the parameters
            will be treated as a half-cell setup with said electrode.
        """

        from ep_bolfi.models.standard_parameters import (
            cₙ, cₚ, SOCₙ_init, SOCₚ_init,
            iₛₑₙ_0, iₛₑₚ_0,
            zₙ, zₚ,
            αₙₙ, αₚₙ, αₙₚ, αₚₚ,
            σₙ, σₚ,
            cₑ_init,
            εₛ_scalar, βₑₛ_scalar,
            Dₑ, κₑ, κₑ_hat, t_plus, one_plus_dlnf_dlnc,
            Lₙ_dim, Lₚ_dim, Lₛ,
            Cₑ, Cᵣₙ, Cᵣₚ,
            γₑ, γₙ, γₚ,
            A, F,
            T_init, thermal_voltage,
        )

        ## The parameter dictionary converted to PyBaMM form.
        # Convert SubstitutionDict to dict by iterating over it.
        self.parameters = pybamm.ParameterValues({
            k: v for k, v in parameters.items()
        })
        ## Sets whether full- or half-cell parameters are calculated.
        self.working_electrode = working_electrode
        if self.working_electrode == 'positive':
            Lₙ = pybamm.Scalar(0)
            εₙ_scalar = pybamm.Scalar(0)
            βₑₙ_scalar = pybamm.Scalar(0)
            βₛₙ_scalar = pybamm.Scalar(1)
            from ep_bolfi.models.standard_parameters import (
                Lₚ, εₚ_scalar, βₑₚ_scalar, βₛₚ_scalar
            )
        elif self.working_electrode == 'negative':
            Lₚ = pybamm.Scalar(0)
            εₚ_scalar = pybamm.Scalar(0)
            βₑₚ_scalar = pybamm.Scalar(0)
            βₛₚ_scalar = pybamm.Scalar(1)
            from ep_bolfi.models.standard_parameters import (
                Lₙ, εₙ_scalar, βₑₙ_scalar, βₛₙ_scalar
            )
        else:
            from ep_bolfi.models.standard_parameters import (
                Lₙ, Lₚ, εₙ_scalar, εₚ_scalar,
                βₑₙ_scalar, βₛₙ_scalar, βₑₚ_scalar, βₛₚ_scalar
            )

        ## Negative electrode theoretical capacity.
        self.Qₙ = (1 - εₙ_scalar) * Lₙ_dim * cₙ * zₙ * F * A
        ## Positive electrode theoretical capacity.
        self.Qₚ = (1 - εₚ_scalar) * Lₚ_dim * cₚ * zₚ * F * A

        ## Integration constant for the electrolyte concentration / I.
        self.R_cₑ_0_1 = (
            (1 - t_plus(cₑ_init)) / (6 * Dₑ(cₑ_init, T_init) * γₑ) * (
                2 * Lₚ**2 / εₚ_scalar**βₑₚ_scalar
                - 2 * Lₙ**2 / εₙ_scalar**βₑₙ_scalar
                + 3 * (Lₙ**2 - Lₚ**2 + 1) / εₛ_scalar**βₑₛ_scalar
            )
        )
        ## Electrolyte concentration / I at the negative electrode.
        self.R_bar_cₑₙ_1 = (
            (1 - t_plus(cₑ_init)) / (6 * Dₑ(cₑ_init, T_init) * γₑ) * (
                2 * Lₙ / εₙ_scalar**βₑₙ_scalar
                - 6 * Lₙ / εₛ_scalar**βₑₛ_scalar
            ) + self.R_cₑ_0_1
        )
        ## Electrolyte concentration / I at the positive electrode.
        self.R_bar_cₑₚ_1 = (
            (1 - t_plus(cₑ_init)) / (6 * Dₑ(cₑ_init, T_init) * γₑ) * (
                -2 * Lₚ / εₚ_scalar**βₑₚ_scalar
                - 6 * (1 - Lₚ) / εₛ_scalar**βₑₛ_scalar
            ) + self.R_cₑ_0_1
        )
        if self.working_electrode == 'positive':
            ## Effective resistance of the negative electrode.
            self.R_bar_φₛₙ_1 = -Lₙ / (σₙ * Cₑ)
        else:
            self.R_bar_φₛₙ_1 = -Lₙ / (
                3 * σₙ * (1 - εₙ_scalar)**βₛₙ_scalar * Cₑ
            )
        if self.working_electrode == 'negative':
            ## Effective resistance of the positive electrode.
            self.R_bar_φₛₚ_1 = Lₚ / (
                3 * σₚ * (1 - εₚ_scalar)**βₛₚ_scalar * Cₑ
            )
        else:
            self.R_bar_φₛₚ_1 = Lₚ / (3 * σₚ * (1 - εₚ_scalar)**βₛₚ_scalar * Cₑ)

        ## Negative electrode resistance (as calculated by the SPMe(S)).
        self.Rₙₛ = self.eval(thermal_voltage) * self.eval(
            -Lₙ / (3 * (1 - εₙ_scalar)**βₛₙ_scalar * σₙ)
        )
        ## Positive electrode resistance (as calculated by the SPMe(S)).
        self.Rₚₛ = self.eval(thermal_voltage) * self.eval(
            -Lₚ / (3 * (1 - εₚ_scalar)**βₛₚ_scalar * σₚ)
        )
        ## Electrolyte resistance (as calculated by the SPMe(S)).
        self.Rₑ = self.eval(thermal_voltage * Cₑ) * self.eval(
            (2 * (1 - t_plus(cₑ_init)) * one_plus_dlnf_dlnc(cₑ_init))
            * (self.R_bar_cₑₚ_1 - self.R_bar_cₑₙ_1)
            + self.R_bar_cₑₚ_1 / zₚ - self.R_bar_cₑₙ_1 / zₙ - (
                Lₚ / (3 * εₚ_scalar**βₑₚ_scalar)
                + Lₙ / (3 * εₙ_scalar**βₑₙ_scalar)
                + Lₛ / εₛ_scalar**βₑₛ_scalar
            ) / (κₑ(cₑ_init, T_init) * Cₑ * κₑ_hat)
        )

        ## SPM(e) negative electrode exchange-current.
        self.iₛₑₙ = lambda ηₙ: (
            self.eval(
                (γₙ / Cᵣₙ) * iₛₑₙ_0(cₑ_init, SOCₙ_init(0), cₙ, T_init)
            ) * (
                np.exp(self.eval(αₙₙ * zₙ) * ηₙ)
                - np.exp(self.eval(-αₚₙ * zₙ) * ηₙ)
            )
        )
        ## SPM(e) positive electrode exchange-current.
        self.iₛₑₚ = lambda ηₚ: (
            self.eval(
                (γₚ / Cᵣₚ) * iₛₑₚ_0(cₑ_init, SOCₚ_init(1), cₚ, T_init)
            ) * (
                np.exp(self.eval(αₙₚ * zₚ) * ηₚ)
                - np.exp(self.eval(-αₚₚ * zₚ) * ηₚ)
            )
        )

    def eval(self, expression):
        """!@brief Short-hand for PyBaMM symbol evaluation.
        @param expression
            A pybamm.Symbol.
        @return
            The numeric value of "expression".
        """
        return self.parameters.process_symbol(expression).evaluate()

    def print(self, c_rates=[0.1, 0.2, 0.5, 1.0]):
        """!@brief Prints the voltage losses for the given C-rates.
        @param c_rates
            The C-rates (as fraction of "Typical current [A]").
        """

        from ep_bolfi.models.standard_parameters import (
            SOCₙ_init, SOCₚ_init,
            iₛₑₙ_0, d_cₑₙ_iₛₑₙ_0, iₛₑₚ_0, d_cₑₚ_iₛₑₚ_0,
            zₙ, zₚ,
            αₙₙ, αₚₙ, αₙₚ, αₚₚ,
            cₑ_init, cₙ, cₚ,
            Lₙ, Lₚ,
            Cₑ,
            T_init, thermal_voltage,
        )

        c_rates = np.array(c_rates)

        if self.working_electrode == 'positive':
            Lₙ_inverse = 1
            Lₚ_inverse = self.eval(1 / Lₚ)
        elif self.working_electrode == 'negative':
            Lₙ_inverse = self.eval(1 / Lₙ)
            Lₚ_inverse = 1
        else:
            Lₙ_inverse = self.eval(1 / Lₙ)
            Lₚ_inverse = self.eval(1 / Lₚ)
        # SPM negative electrode overpotential.
        ηₙ_0 = np.array([root_scalar(
            lambda ηₙ: self.iₛₑₙ(ηₙ) - c * Lₙ_inverse,
            method='toms748', bracket=[-40, 40], x0=0
        ).root for c in c_rates])
        # SPM positive electrode overpotential.
        ηₚ_0 = np.array([root_scalar(
            lambda ηₚ: self.iₛₑₚ(ηₚ) + c * Lₚ_inverse,
            method='toms748', bracket=[-40, 40], x0=0
        ).root for c in c_rates])
        # SPMe negative electrolyte concentration.
        bar_cₑₙ_1 = self.eval(self.R_bar_cₑₙ_1) * c_rates
        # SPMe positive electrolyte concentration.
        bar_cₑₚ_1 = self.eval(self.R_bar_cₑₚ_1) * c_rates
        # SPMe negative electrode overpotential correction.
        bar_ηₙ_1 = (
            -bar_cₑₙ_1 / self.eval(zₙ)
            * self.eval(d_cₑₙ_iₛₑₙ_0(cₑ_init, SOCₙ_init(0), cₙ, T_init)
                        / iₛₑₙ_0(cₑ_init, SOCₙ_init(0), cₙ, T_init))
            * (np.exp(self.eval(αₙₙ * zₙ) * ηₙ_0)
               - np.exp(self.eval(-αₚₙ * zₙ) * ηₙ_0))
            / (self.eval(αₙₙ) * np.exp(self.eval(αₙₙ * zₙ) * ηₙ_0)
               + self.eval(αₚₙ) * np.exp(self.eval(-αₚₙ * zₙ) * ηₙ_0))
        )
        # SPMe positive electrode overpotential correction.
        bar_ηₚ_1 = (
            -bar_cₑₚ_1 / self.eval(zₚ)
            * self.eval(d_cₑₚ_iₛₑₚ_0(cₑ_init, SOCₚ_init(1), cₚ, T_init)
                        / iₛₑₚ_0(cₑ_init, SOCₚ_init(1), cₚ, T_init))
            * (np.exp(self.eval(αₙₚ * zₚ) * ηₚ_0)
               - np.exp(self.eval(-αₚₚ * zₚ) * ηₚ_0))
            / (self.eval(αₙₚ) * np.exp(self.eval(αₙₚ * zₚ) * ηₚ_0)
               + self.eval(αₚₚ) * np.exp(self.eval(-αₚₚ * zₚ) * ηₚ_0))
        )
        # SPM electrolyte potential.
        # if self.halfcell:
        #     φₑ_0 = -ηₙ_0
        # else:
        #     φₑ_0 = -ηₙ_0 - self.eval(OCVₙ(SOCₙ_init(0), T_init))
        # SPMe electrolyte potential integration constant
        # from ep_bolfi.models.standard_parameters import (
        #     σₙ, εₙ_scalar, εₛ_scalar, βₑₙ_scalar, βₑₛ_scalar, βₛₙ_scalar,
        #     κₑ, κₑ_hat, t_plus, one_plus_dlnf_dlnc,
        # )
        # if self.working_electrode == 'positive':
        #     tilde_φₑ = (
        #         -bar_ηₙ_1 - bar_cₑₙ_1 / self.eval(zₙ)
        #         - self.eval(2 * (1 - t_plus(cₑ_init))
        #                     * one_plus_dlnf_dlnc(cₑ_init)) * bar_cₑₙ_1
        #     )
        # else:
        #     tilde_φₑ = (
        #         -c_rates * self.eval(
        #             Lₙ / (3 * σₙ * (1 - εₙ_scalar)**βₛₙ_scalar * Cₑ)
        #         ) - bar_ηₙ_1 - bar_cₑₙ_1 / self.eval(zₙ)
        #         - self.eval(2 * (1 - t_plus(cₑ_init))
        #                     * one_plus_dlnf_dlnc(cₑ_init)) * bar_cₑₙ_1
        #         - c_rates / self.eval(κₑ_hat * Cₑ * κₑ(cₑ_init, T_init))
        #         * self.eval((-Lₙ / (3 * εₙ_scalar**βₑₙ_scalar)
        #                     + Lₙ / εₛ_scalar**βₑₛ_scalar))
        #     )

        # SPMe electrolyte potential correction
        # bar_φₑₙ_1 = tilde_φₑ + self.eval(
        #     2 * (1 - t_plus(cₑ_init)) * one_plus_dlnf_dlnc(cₑ_init)
        # ) * bar_cₑₙ_1 - c_rates / self.eval(
        #     κₑ_hat * Cₑ * κₑ(cₑ_init, T_init)
        # ) * (
        #     self.eval(-Lₙ / (3 * εₙ_scalar**βₑₙ_scalar)
        #               + Lₙ / εₛ_scalar**βₑₛ_scalar)
        # )
        # bar_φₑₚ_1 = tilde_φₑ + self.eval(
        #     2 * (1 - t_plus(cₑ_init)) * one_plus_dlnf_dlnc(cₑ_init)
        # ) * bar_cₑₚ_1 - c_rates / self.eval(
        #     κₑ_hat * Cₑ * κₑ(cₑ_init, T_init)
        # ) * (
        #     self.eval(Lₚ / (3 * εₚ_scalar**βₑₚ_scalar)
        #               + (1 - Lₚ) / εₛ_scalar**βₑₛ_scalar)
        # )

        # Tabulate the voltage losses.
        table = [
            ["Current / C-rate"] + list(c_rates),
            ["SPM anode OP [mV]"] + list(self.eval(thermal_voltage) * ηₙ_0),
            ["SPM cathode OP [mV]"] + list(-self.eval(thermal_voltage) * ηₚ_0),
            ["SPMe anode OP [mV]"] + list(self.eval(thermal_voltage) * (
                ηₙ_0 - self.eval(Cₑ) * bar_ηₙ_1
            )),
            ["SPMe cathode OP [mV]"] + list(-self.eval(thermal_voltage) * (
                ηₚ_0 + self.eval(Cₑ) * bar_ηₚ_1
            )),
            ["SPMe electrolyte drop [mV]"] + list(-self.Rₑ * c_rates),
            ["SPMe anode drop [mV]"] + list(-self.Rₙₛ * c_rates),
            ["SPMe cathode drop [mV]"] + list(-self.Rₚₛ * c_rates),
        ]
        for row in range(1, len(table)):
            for column in range(1, len(table[row])):
                table[row][column] = '{:.3f}'.format(table[row][column] * 1e3)

        print("Negative electrode theoretical capacity: "
              + str(self.eval(self.Qₙ)) + " C")
        print("Positive electrode theoretical capacity: "
              + str(self.eval(self.Qₚ)) + " C")
        print(tabulate(table, headers='firstrow', tablefmt='pretty'))



"""!@package ep_bolfi.utility.fitting_functions
Various helper and fitting functions for processing measurement curves.
"""

import json
import re

import numpy as np
from pyimpspec import calculate_drt_tr_nnls, DataSet, TRNNLSResult
import scipy.optimize as so
from scipy.optimize import root_scalar, minimize
from scipy import interpolate as ip
from sklearn.cluster import KMeans
import sympy as sp


class NDArrayEncoder(json.JSONEncoder):
    def default(self, item):
        if isinstance(item, np.ndarray):
            return item.tolist()
        return json.JSONEncoder.default(self, item)


class OCV_fit_result(object):
    """!@brief Contains OCV fit parameters and related information.
    Reference
    ----------
    C. R. Birkl, E. McTurk, M. R. Roberts, P. G. Bruce and D. A. Howey.
    “A Parametric Open Circuit Voltage Model for Lithium Ion Batteries”.
    Journal of The Electrochemical Society, 162(12):A2271-A2280, 2015
    """

    def __init__(
        self,
        fit,
        SOC,
        OCV,
        SOC_offset=1.0,
        SOC_scale=1.0,
        optimize_result=None,
        spline_interpolation_knots=None,
        spline_interpolation_coefficients=None,
        function_string=None,
    ):
        if len(fit) % 3:
            ## The SOC range of the data.
            self.SOC_range = np.array(fit[0:2])
            ## The fit parameters of the OCV function from Birkl. et al.,
            # excluding the estimated SOC range.
            self.fit = np.array(fit[2:])
        else:
            self.SOC_range = np.array([0, 1])
            self.fit = np.array(fit)
        ## The E₀ (plateau voltages) parameters.
        self.E_0 = np.array(self.fit[0::3])
        ## The a (inverse plateau widths) parameters.
        self.a = np.array(self.fit[1::3])
        ## The Δx (phase proportion) parameters.
        self.Δx = np.array(self.fit[2::3])
        if len(self.Δx) < len(self.E_0):
            last_Δx = 1 - np.sum(self.Δx)
            self.Δx = np.append(self.Δx, last_Δx)
            self.fit = np.append(self.fit, last_Δx)
        self.E_0, self.a, self.Δx = zip(*sorted(zip(
            self.E_0, self.a, self.Δx
        ), reverse=True))
        self.fit = [
            par
            for plateau in zip(self.E_0, self.a, self.Δx)
            for par in plateau
        ]
        for i in range(len(self.E_0) - 1):
            if (
                np.abs(self.E_0[i + 1] - self.E_0[i])
                + np.abs(self.a[i + 1] - self.a[i]) < 1e-4
            ):
                print("Warning (OCV_fit_result): At least two fitted summands "
                      "coincide.")
        ## The SOC data points.
        self.SOC = np.array(SOC)
        ## The OCV data points. May be adjusted from the original data.
        self.OCV = np.array(OCV)
        ## If another electrode was factored out in the data, this may
        # contain its SOC at SOC 0 of the electrode of interest.
        self.SOC_offset = SOC_offset
        ## If another electrode was factored out in the data, this may
        # contain the the rate of change of its SOC to that of the
        # electrode of interest.
        self.SOC_scale = SOC_scale
        ## The scipy.optimize.OptimizeResult that led to the fit.
        self.optimize_result = optimize_result
        ## The knots of the interpolating spline fitted to the inverse.
        self.spline_interpolation_knots = spline_interpolation_knots
        ## The coefficients of the interpolating spline fitted to the inverse.
        self.spline_interpolation_coefficients = (
            spline_interpolation_coefficients
        )
        ## The string representation of the interpolating spline fitted to the
        # inverse.
        self.function_string = function_string

    def to_json(self):
        return json.dumps({
            'fit': {
                'parameters': self.fit,
                'SOC_range': self.SOC_range,
                'E_0': self.E_0,
                'a': self.a,
                'Δx': self.Δx,
            },
            'data': {
                'SOC': self.SOC,
                'OCV': self.OCV,
            },
            'balancing': {
                'SOC_offset': self.SOC_offset,
                'SOC_scale': self.SOC_scale,
            },
            'optimize_result': self.optimize_result,
            'spline': {
                'knots': self.spline_interpolation_knots,
                'coefficients': self.spline_interpolation_coefficients,
            },
            'function': self.function_string,
        }, cls=NDArrayEncoder)

    def SOC_adjusted(self, soc=None):
        """!@brief Gives the adjusted SOC values.
        @param soc
            The SOC as assigned in the original data. This usually
            corresponds to the range available during a measurement.
        @return
            The SOC as corrected by the OCV model. These values will try
            to correspond to the level of lithiation.
        """

        if soc is None:
            soc = self.SOC

        return self.SOC_range[0] + soc * (
            self.SOC_range[1] - self.SOC_range[0]
        )

    def SOC_other_electrode(self, soc=None):
        """!@brief Relates the SOCs of the two electrodes to each other.
        If the original data was of a full cell and the other
        electrode was factored out, this may contain the function that
        takes the SOC of the electrode of interest and gives the SOC
        of the other electrode, i.e., the stoichiometric relation.
        @param soc
            The SOC of the electrode of interest.
        @return
            The SOC of the other electrode that was factored out.
        """

        if soc is None:
            soc = self.SOC

        return self.SOC_offset - self.SOC_scale * soc


def find_occurrences(sequence, value):
    """!@brief Gives indices in sequence where it is closest to value.
    @param sequence
        A list that represents a differentiable function.
    @param value
        The value that is searched for in "sequence". Also, crossings of
        consecutive values in "sequence" with "value" are searched for.
    @return
        A list of indices in "sequence" in ascending order where "value"
        or a close match for "value" was found.
    """
    root_finder = np.array(sequence) - value
    crossings = (
        np.sign(root_finder[1:]) - np.sign(root_finder[:-1])
    ).nonzero()[0]
    nearest_indices = np.where(
        np.abs(root_finder[crossings]) < np.abs(root_finder[crossings + 1]),
        crossings, crossings + 1
    )
    if len(nearest_indices) == 0:
        nearest_indices = (
            [0]
            if np.abs(root_finder[0]) < np.abs(root_finder[-1]) else
            [len(sequence) - 1]
        )
    return nearest_indices


def smooth_fit(
    x, y, order=3, splits=None, w=None, s=None, display=False, derivatives=0
):
    """!@brief Calculates a smoothed spline with derivatives.
    Note: the "roots" function of a spline only works if it is cubic,
    i.e. of third order. Each "derivative" reduces the order by one.
    @param x
        The independent variable.
    @param y
        The dependent variable ("plotted over x").
    @param order
        Interpolation order of the spline.
    @param splits
        Optional tuning parameter. A list of points between which
        splines will be fitted first. The returned spline then is a fit
        of these individual splines.
    @param w
        Optional list of weights. Works best if 1 / w approximates the
        standard deviation of the noise in y at each point. Defaults to
        SciPy default behaviour of scipy.interpolate.UnivariateSpline.
    @param s
        Optional tuning parameter. Higher values lead to coarser, but
        more smooth interpolations and vice versa. Defaults to SciPy
        default behaviour of scipy.interpolate.UnivariateSpline.
    @param display
        If set to True, the fit parameters of the spline will be printed
        to console. If possible, a monomial representation is printed.
    @param derivatives
        The derivatives of the spline to also include in the return.
        Default is 0, which gives the spline. 1 would give the spline,
        followed by its derivative. Can not be higher than spline order.
        Derivatives are only continuous when derivatives < order.
    @return
        A smoothing spline in the form of scipy.UnivariateSpline.
    """

    # Clean up multiple-valued ranges.
    x = np.array(x)
    y = np.array(y)

    def recursive_cleanup(x, y):
        increasing = np.append(True, x[1:] > x[:-1])
        # stationary = np.append(True, x[1:] == x[:-1])
        decreasing = np.append(True, x[1:] < x[:-1])
        if np.sum(increasing) == 1 or np.sum(decreasing) == 1:
            return x, y
        # Choose the larger range of monotonic x values.
        if np.sum(increasing) > np.sum(decreasing):
            x = x[increasing]
            y = y[increasing]
        else:
            # Make sure that x is increasing instead of decreasing.
            x = np.flip(x[decreasing])
            y = np.flip(y[decreasing])
        x, y = recursive_cleanup(x, y)
        return x, y
    x, y = recursive_cleanup(x, y)

    y_smoothed = []

    # One could use the smoothing spline for the data directly, but that
    # isn't as stable as subdividing the data first in some cases.
    # One such case is if there are high variations in gradients.
    if splits is None:
        y_smoothed = y
    else:
        split_indices = [np.abs(x-split).argmin() for split in splits]
        x_splits = np.split(x, split_indices)
        number_of_splits = len(x_splits)
        y_splits = np.split(y, split_indices)

        for x_split, y_split in zip(x_splits, y_splits):
            if len(x_split) < order-2:
                print("Warning: not enough points in split for smoothing.")
                y_smoothed.extend(y_split)
            else:
                split_spline = ip.UnivariateSpline(
                    x_split, y_split, w=w, k=order,
                    s=None if s is None else s/number_of_splits
                )
                y_smoothed.extend(split_spline(x_split))

    try:
        spline = ip.UnivariateSpline(x, y_smoothed,
                                     k=order, s=s)
    except ValueError:
        spline = ip.UnivariateSpline(x[::-1], y_smoothed[::-1],
                                     k=order, s=s)

    if display:
        print("Knots of the interpolating spline:")
        print(spline.get_knots())
        print("Coefficients of the interpolating spline:")
        print(spline.get_coeffs())
        try:
            simplification = verbose_spline_parameterization(
                spline.get_coeffs(), spline.get_knots(), order,
                function_name="y", function_args="x", derivatives=derivatives,
                verbose=True
            )
            print("Monomial representation of the interpolating spline:")
            print(simplification)
        except ValueError as e:
            print(e)

    return spline


def fit_exponential_decay(
    timepoints, voltages, recursive_depth=1, threshold=0.95
):
    """!@brief Extracts a set amount of exponential decay curves.
    @param timepoints
        The timepoints of the measurements.
    @param voltages
        The corresponding voltages.
    @param recursive_depth
        The default (1) fits one exponential curve to the data. For
        higher values that fit is repeated with the data minus the
        preceding fit(s) for this amount of times minus one.
    @param threshold
        The lower threshold value for the R² coefficient of
        determination. If "threshold" is smaller than 1, the subset of
        the exponential decay data is searched that just fulfills it.
        Defaults to 0.95. Values above 1 are set to 1.
    @return
        A list of length "recursive_depth" where each element is a
        3-tuple with the timepoints, the fitted voltage evaluations
        and a 3-tuple of the parameters of the following decay function:
        t, (U_0, ΔU, τᵣ⁻¹) ↦ U_0 + ΔU * np.exp(-τᵣ⁻¹ * (t - timepoints[0])).
    """

    t_eval = np.atleast_1d(timepoints)
    u_eval = np.atleast_1d(voltages)
    threshold = threshold if threshold <= 1 else 1

    # Make sure that invalid inputs don't crash anything.
    if len(t_eval) < 3:
        print("Warning: fit_exponential_decay was given insufficient data.")
        return []

    def exp_fit_function(t, b, c, d, t_0=t_eval[0]):
        """! Exponential decay function. """
        return b + c * np.exp(-d * (t - t_0))

    def log_inverse_fit_function(y, b, c, d, t_0=t_eval[0]):
        """! Corresponding logarithm function. """
        log_arg = (y - b) / c
        log_arg[log_arg <= 0] = 0.1**d
        return t_0 - np.log(log_arg) / d

    end = len(t_eval) - 1
    bracket = [0, end]
    curves = []
    depth_counter = 0
    fit_guess = [u_eval[end], u_eval[end // 10] - u_eval[end],
                 1.0 / (t_eval[end] - t_eval[end // 10])]

    # Evaluate the R² value for a split at the middle of the data.
    split = int(0.5 * (bracket[0] + bracket[1]))
    argmax_R_squared = split
    fit_split = so.minimize(
        lambda x: np.sum(
            (exp_fit_function(t_eval[split:end], *x) - u_eval[split:end])**2
        )**0.5, x0=fit_guess, method='trust-constr'
    ).x
    test_t_split = log_inverse_fit_function(u_eval[split:end], *fit_split)
    R_squared_split = (
        1 - np.sum(np.nan_to_num(t_eval[split:end] - test_t_split)**2)
        / np.sum((t_eval[split:end] - np.mean(t_eval[split:end]))**2)
    )

    while True:
        # End prematurely if the threshold is matched.
        if R_squared_split >= threshold:
            fit_eval = exp_fit_function(t_eval, *fit_split)
            R_squared_split = -float('inf')
            end = split
            bracket = [0, end]
            curves.append([t_eval, list(fit_eval), fit_split])
            u_eval = u_eval - fit_eval + fit_eval[0]
            depth_counter = depth_counter + 1
            if depth_counter >= recursive_depth:
                break
            fit_guess = [
                u_eval[end], u_eval[end // 10] - u_eval[end],
                1.0 / (t_eval[end] - t_eval[end // 10])
            ]
        # If the threshold wasn't reached, use the highest R² value.
        if bracket[1] - bracket[0] <= 1:
            fit_argmax = so.minimize(lambda x: np.sum(
                (exp_fit_function(t_eval[argmax_R_squared:end], *x)
                 - u_eval[argmax_R_squared:end])**2
            )**0.5, x0=fit_guess, method='trust-constr').x
            fit_eval = exp_fit_function(t_eval, *fit_argmax)
            end = split
            bracket = [0, end]
            curves.append([t_eval, list(fit_eval), fit_argmax])
            u_eval = u_eval - fit_eval + fit_eval[0]
            depth_counter = depth_counter + 1
            if depth_counter >= recursive_depth:
                break
            fit_guess = [
                u_eval[end], u_eval[end // 10] - u_eval[end],
                1.0 / (t_eval[end] - t_eval[end // 10])
            ]
        left = int(0.75 * bracket[0] + 0.25 * bracket[1])
        split = int(0.5 * (bracket[0] + bracket[1]))
        right = int(0.25 * bracket[0] + 0.75 * bracket[1])
        # Fit an exponential decay for both "left" and "right" splits.
        fit_left = so.minimize(lambda x: np.sum(
            (exp_fit_function(t_eval[left:end], *x)
             - u_eval[left:end])**2
        )**0.5, x0=fit_guess, method='trust-constr').x
        fit_right = so.minimize(lambda x: np.sum(
            (exp_fit_function(t_eval[right:end], *x)
             - u_eval[right:end])**2
        )**0.5, x0=fit_guess, method='trust-constr').x
        # Take the logarithm corresponding to the fit. If the fit is
        # good, it should give a line, and thus, a high R² value.
        test_t_left = log_inverse_fit_function(u_eval[left:end], *fit_left)
        R_squared_left = (
            1 - np.sum(np.nan_to_num(t_eval[left:end] - test_t_left)**2)
            / np.sum((t_eval[left:end] - np.mean(t_eval[left:end]))**2)
        )
        test_t_right = log_inverse_fit_function(u_eval[right:end], *fit_right)
        R_squared_right = (
            1 - np.sum(np.nan_to_num(t_eval[right:end] - test_t_right)**2)
            / np.sum((t_eval[right:end] - np.mean(t_eval[right:end]))**2)
        )
        # Step in the direction of ascending R² value.
        if R_squared_left > R_squared_right:
            bracket[1] = split
            if R_squared_left > R_squared_split:
                argmax_R_squared = left
            R_squared_split = R_squared_left
            fit_split = fit_left
        else:
            bracket[0] = split
            if R_squared_right > R_squared_split:
                argmax_R_squared = right
            R_squared_split = R_squared_right
            fit_split = fit_right

    return curves


def fit_sqrt(timepoints, voltages, threshold=0.95):
    """!@brief Extracts a square root at the beginning of the data.
    @param timepoints
        The timepoints of the measurements.
    @param voltages
        The corresponding voltages.
    @param threshold
        The lower threshold value for the R² coefficient of
        determination. If "threshold" is smaller than 1, the subset of
        the experimental data is searched that just fulfills it.
        Defaults to 0.95. Values above 1 are set to 1.
    @return
        A 3-tuple with the timepoints, the fitted voltage evaluations
        and a 2-tuple of the parameters of the following sqrt function:
        t, (U_0, dU_d√t) ↦ U_0 + dU_d√t * √(t - timepoints[0]).
    """

    t_eval = np.atleast_1d(timepoints)
    u_eval = np.atleast_1d(voltages)
    threshold = threshold if threshold <= 1 else 1

    # Make sure that invalid inputs don't crash anything.
    if len(t_eval) < 2:
        print("Warning: fit_sqrt was given insufficient data.")
        return []

    def sqrt_fit_function(t, b, c, t_0=t_eval[0]):
        """! Square root function. """
        return b + c * np.sqrt(t - t_0)

    def square_inverse_fit_function(y, b, c, t_0=t_eval[0]):
        """! Corresponding square function. """
        return t_0 + ((y - b) / c)**2

    end = len(t_eval) - 1
    bracket = [0, end]
    fit_guess = [u_eval[0],
                 (u_eval[end] - u_eval[0]) / np.sqrt(t_eval[end] - t_eval[0])]

    # Evaluate the R² value for a split at the middle of the data.
    split = int(0.5 * (bracket[0] + bracket[1]))
    argmax_R_squared = split
    fit_split = so.minimize(lambda x: np.sum(
        (sqrt_fit_function(t_eval[0:split], *x) - u_eval[0:split])**2
    )**0.5, x0=fit_guess, method='trust-constr').x
    test_t_split = square_inverse_fit_function(u_eval[0:split], *fit_split)
    R_squared_split = (
        1 - np.sum(np.nan_to_num(t_eval[0:split] - test_t_split)**2)
        / np.sum((t_eval[0:split] - np.mean(t_eval[0:split]))**2)
    )

    while True:
        # Return the result if the threshold is matched.
        if R_squared_split >= threshold:
            fit_eval = sqrt_fit_function(t_eval, *fit_split)
            return [t_eval, list(fit_eval), fit_split]
        # If the threshold wasn't reached, use the highest R² value.
        if bracket[1] - bracket[0] <= 1:
            fit_argmax = so.minimize(
                lambda x: np.sum(
                    (sqrt_fit_function(t_eval[0:argmax_R_squared], *x)
                     - u_eval[0:argmax_R_squared])**2
                )**0.5, x0=fit_guess, method='trust-constr'
            ).x
            fit_eval = sqrt_fit_function(t_eval, *fit_argmax)
            return [t_eval, list(fit_eval), fit_argmax]
        left = int(0.75 * bracket[0] + 0.25 * bracket[1])
        split = int(0.5 * (bracket[0] + bracket[1]))
        right = int(0.25 * bracket[0] + 0.75 * bracket[1])
        # Fit an exponential decay for both "left" and "right" splits.
        fit_left = so.minimize(lambda x: np.sum(
            (sqrt_fit_function(t_eval[0:left], *x) - u_eval[0:left])**2
        )**0.5, x0=fit_guess, method='trust-constr').x
        fit_right = so.minimize(lambda x: np.sum(
            (sqrt_fit_function(t_eval[0:right], *x) - u_eval[0:right])**2
        )**0.5, x0=fit_guess, method='trust-constr').x
        # Take the square corresponding to the fit. If the fit is
        # good, it should give a line, and thus, a high R² value.
        test_t_left = square_inverse_fit_function(u_eval[0:left], *fit_left)
        R_squared_left = (
            1 - np.sum(np.nan_to_num(t_eval[0:left] - test_t_left)**2)
            / np.sum((t_eval[0:left] - np.mean(t_eval[0:left]))**2)
        )
        test_t_right = square_inverse_fit_function(u_eval[0:right], *fit_right)
        R_squared_right = (
            1 - np.sum(np.nan_to_num(t_eval[0:right] - test_t_right)**2)
            / np.sum((t_eval[0:right] - np.mean(t_eval[0:right]))**2)
        )
        # Step in the direction of ascending R² value.
        if R_squared_left > R_squared_right:
            bracket[1] = split
            if R_squared_left > R_squared_split:
                argmax_R_squared = left
            R_squared_split = R_squared_left
            fit_split = fit_left
        else:
            bracket[0] = split
            if R_squared_right > R_squared_split:
                argmax_R_squared = right
            R_squared_split = R_squared_right
            fit_split = fit_right


def fit_drt(frequencies, impedances, lambda_value=-2.0):
    """Distribution of Relaxation Times.
    @param frequencies
        Array of the measured frequencies.
    @param impedances
        Array of the measured complex impedances.
    @param lambda_value
        Takes the place of the R² value as a tuning parameter.
        Consult the pyimpspec documentation for the precise usage.
        Default is -2, which "uses the L-curve approach to estimate lambda".
        -1 uses a different heuristic, and values > 0 set lambda directly.
    @return
        A 3-tuple of characteristic DRT time constants, their corresponding
        resistances, and the whole TRNNLSResult object returned by pyimpspec.
        Note that the .pseudo_chisqr attribute is not useful for the same
        reasons that went into the R² calculations in the other fit functions.
    """

    # Clean up any "inductive" impedances, i.e., those with imag < 0.
    positive_indices = np.imag(impedances) < 0
    frequencies = np.array(frequencies)[positive_indices]
    impedances = np.array(impedances)[positive_indices]
    pyimpspec_dataset = DataSet(frequencies, impedances)
    drt: TRNNLSResult = calculate_drt_tr_nnls(
        pyimpspec_dataset, lambda_value=lambda_value
    )
    # Assume that the peaks are sharp. So the adjacent discrete time constants
    # have all the weight of every peak (rest of the gammas should be zero).
    # Since Z = R_0 + ∫₀∞ gamma / (1 + s w) dw, the actual resistances are the
    # integrals around the peaks.
    # pyimpspec uses a logarithmic transform for the integral, i.e.,
    # the "dw" is actually a "d(log(w))".
    peaks = drt.get_peaks(threshold=0.0)
    peak_indices = [
        find_occurrences(drt.time_constants, time_constant)[0]
        for time_constant in peaks[0]
    ]
    # peak_values = peaks[1]
    resistances = []
    for pi in peak_indices:
        peak_surrounding = np.array(
            [0] + list(drt.gammas[pi - 1: pi + 2]) + [0]
        )
        extended_time_constants = (
            [np.exp(
                2 * np.log(drt.time_constants[0])
                - np.log(drt.time_constants[1])
            )]
            + list(drt.time_constants)
            + [np.exp(
                2 * np.log(drt.time_constants[-1])
                - np.log(drt.time_constants[-2])
            )]
        )
        log_time_constants = np.log(extended_time_constants)
        # Count from one index further, since the frond got appended to.
        integration_weights = np.diff(log_time_constants[pi - 1: pi + 4])
        peak_integral = np.sum(
            0.5 * (peak_surrounding[1:] + peak_surrounding[:-1])
            * integration_weights
        )
        resistances.append(peak_integral)
    return (drt.time_constants[peak_indices], resistances, drt)


def laplace_transform(x, y, s):
    """!@brief Performs a basic laplace transformation.
    @param x
        The independent variable.
    @param y
        The dependent variable.
    @param s
        The (possibly complex) frequencies for which to perform the
        transform.
    @return
        The evaluation of the laplace transform at s.
    """

    x = np.array(x)
    y = np.array(y)
    Δx = x[1:] - x[:-1]
    x_int = 0.5 * (x[1:] + x[:-1])
    y_int = 0.5 * (y[1:] + y[:-1])
    return np.sum(y_int * np.exp(-(s[:, None] + 0.0j) * x_int) * Δx, axis=1)


def a_fit(γUeminus1):
    """!@brief Calculates the conversion from "A parametric OCV model".
    @param γUeminus1
        γ * Uᵢ / e from "A parametric OCV model".
    @return
        The approximation factor aᵢ from "A parametric OCV model".
    """

    return 0.0789207 / (γUeminus1 - 0.08048093)


def OCV_fit_function(
    E_OCV,
    *args,
    z=1.0,
    T=298.15,
    individual=False,
    fit_SOC_range=False,
    rescale=False,
):
    """!@brief The OCV model from "A parametric OCV model".
    Reference
    ----------
    C. R. Birkl, E. McTurk, M. R. Roberts, P. G. Bruce and D. A. Howey.
    “A Parametric Open Circuit Voltage Model for Lithium Ion Batteries”.
    Journal of The Electrochemical Society, 162(12):A2271-A2280, 2015

    @param E_OCV
        The voltages for which the SOCs shall be evaluated.
    @param args
        A list which length is dividable by three. These are the
        parameters E₀, a and Δx from the paper referenced above in the
        order (E₀_0, a_0, Δx_0, E₀_1, a_1, Δx_1...). If "fit_SOC_range"
        is True, two additional arguments are at the front (see there).
        The last Δx_j may be omitted to force ∑ⱼ Δx_j = 1.
    @param z
        The charge number of the electrode interface reaction.
    @param T
        The temperature of the electrode.
    @param individual
        If True, the model function summands are not summed up.
    @param fit_SOC_range
        If True, this function takes two additional arguments at the
        start of "args" that may be used to adjust the data SOC range.
    @param rescale
        If True, the expected "args" now contain the slopes of the
        summands at their respective origins instead of a. Formula:
        a / slope = -4 * (k_B / e) * T / (Δx * z).
    @return
        The evaluation of the model function. This is the referenced fit
        function on a linearly transformed SOC range if "fit_SOC_range"
        is True. If "individual" is True, the individual summands in the
        model function get returned as a list.
    """

    if fit_SOC_range:
        offset = 2
        SOC_start = args[0]
        SOC_end = args[1]
    else:
        offset = 0
        SOC_start = 0
        SOC_end = 1

    if (len(args) - offset) % 3 and (len(args) - offset + 1) % 3:
        raise ValueError("OCV_fit_function: incorrect length of args.")

    E_0 = np.array(args[offset + 0::3])
    temp_args = np.array(args[offset + 1::3])
    Δx = np.array(args[offset + 2::3])

    E_OCV = np.array(E_OCV)

    # Ensure that the sum of Δx is 1.
    if len(Δx) < len(E_0):
        Δx = np.append(Δx, 1 - np.sum(Δx))

    e = 1.602176634e-19
    k_B = 1.380649e-23

    # Rescale to obtain a.
    if rescale:
        a = -temp_args * 4 * (k_B / e) * T / (Δx * z)
    else:
        a = temp_args

    try:
        exponential = np.exp((E_OCV[:, None] - E_0) * a * z * e / (k_B * T))
        axis = 1
    except (IndexError, TypeError):
        exponential = np.exp((E_OCV - E_0) * a * z * e / (k_B * T))
        axis = 0
    summands = Δx / (1 + exponential)

    if individual:
        return [np.nan_to_num(SOC_start / len(Δx) + s * (SOC_end - SOC_start))
                for s in summands]
    else:
        return np.nan_to_num(
            SOC_start + np.sum(summands, axis=axis) * (SOC_end - SOC_start)
        )


def d_dE_OCV_fit_function(
    E_OCV,
    *args,
    z=1.0,
    T=298.15,
    individual=False,
    fit_SOC_range=False,
    rescale=False,
):
    """!@brief The derivative of fitting_functions.OCV_fit_function.
    @param E_OCV
        The voltages for which the derivative shall be evaluated.
    @param args
        A list which length is dividable by three. These are the
        parameters E₀, a and Δx from "A parametric OCV model" in the
        order (E₀_0, a_0, Δx_0, E₀_1, a_1, Δx_1...).
        The last Δx_j may be omitted to force ∑ⱼ Δx_j = 1.
    @param z
        The charge number of the electrode interface reaction.
    @param T
        The temperature of the electrode.
    @param individual
        If True, the model function summands are not summed up.
    @param fit_SOC_range
        If True, this function takes two additional arguments at the
        start of "args" that may be used to adjust the data SOC range.
    @param rescale
        If True, the expected "args" now contain the slopes of the
        summands at their respective origins instead of a. Formula:
        a / slope = -4 * (k_B / e) * T / (Δx * z).
    @return
        The evaluation of ∂OCV_fit_function(OCV) / ∂OCV.
    """

    if fit_SOC_range:
        offset = 2
        SOC_start = args[0]
        SOC_end = args[1]
    else:
        offset = 0
        SOC_start = 0
        SOC_end = 1

    if (len(args) - offset) % 3 and (len(args) - offset + 1) % 3:
        raise ValueError("d_dE_OCV_fit_function: incorrect length of args.")

    E_0 = np.array(args[offset + 0::3])
    temp_args = np.array(args[offset + 1::3])
    Δx = np.array(args[offset + 2::3])

    E_OCV = np.array(E_OCV)

    # Ensure that the sum of Δx is 1
    if len(Δx) < len(E_0):
        Δx = np.append(Δx, 1 - np.sum(Δx))

    e = 1.602176634e-19
    k_B = 1.380649e-23

    # Rescale to obtain a.
    if rescale:
        a = -temp_args * 4 * (k_B / e) * T / (Δx * z)
    else:
        a = temp_args

    try:
        exponential = np.exp((E_OCV[:, None] - E_0) * a * z * e / (k_B * T))
        axis = 1
    except (IndexError, TypeError):
        exponential = np.exp((E_OCV - E_0) * a * z * e / (k_B * T))
        axis = 0
    summands = -Δx * a * z * e / (k_B * T) * exponential / (1 + exponential)**2

    if individual:
        return [np.nan_to_num(s * (SOC_end - SOC_start)) for s in summands]
    else:
        return np.nan_to_num(np.sum(summands, axis=axis)) * (
            SOC_end - SOC_start
        )


def d2_dE2_OCV_fit_function(
    E_OCV,
    *args,
    z=1.0,
    T=298.15,
    individual=False,
    fit_SOC_range=False,
    rescale=False,
):
    """!@brief The 2ⁿᵈ derivative of fitting_functions.OCV_fit_function.
    @param E_OCV
        The voltages for which the 2ⁿᵈ derivative shall be evaluated.
    @param args
        A list which length is dividable by three. These are the
        parameters E₀, a and Δx from "A parametric OCV model" in the
        order (E₀_0, a_0, Δx_0, E₀_1, a_1, Δx_1...).
        The last Δx_j may be omitted to force ∑ⱼ Δx_j = 1.
    @param z
        The charge number of the electrode interface reaction.
    @param T
        The temperature of the electrode.
    @param individual
        If True, the model function summands are not summed up.
    @param fit_SOC_range
        If True, this function takes two additional arguments at the
        start of "args" that may be used to adjust the data SOC range.
    @param rescale
        If True, the expected "args" now contain the slopes of the
        summands at their respective origins instead of a. Formula:
        a / slope = -4 * (k_B / e) * T / (Δx * z).
    @return
        The evaluation of ∂²OCV_fit_function(OCV) / ∂²OCV.
    """

    if fit_SOC_range:
        offset = 2
        SOC_start = args[0]
        SOC_end = args[1]
    else:
        offset = 0
        SOC_start = 0
        SOC_end = 1

    if (len(args) - offset) % 3 and (len(args) - offset + 1) % 3:
        raise ValueError("d2_dE2_OCV_fit_function: incorrect length of args.")

    E_0 = np.array(args[offset + 0::3])
    temp_args = np.array(args[offset + 1::3])
    Δx = np.array(args[offset + 2::3])

    E_OCV = np.array(E_OCV)

    # Ensure that the sum of Δx is 1
    if len(Δx) < len(E_0):
        Δx = np.append(Δx, 1 - np.sum(Δx))

    e = 1.602176634e-19
    k_B = 1.380649e-23

    # Rescale to obtain a.
    if rescale:
        a = -temp_args * 4 * (k_B / e) * T / (Δx * z)
    else:
        a = temp_args

    try:
        exponential = np.exp((E_OCV[:, None] - E_0) * a * z * e / (k_B * T))
        axis = 1
    except (IndexError, TypeError):
        exponential = np.exp((E_OCV - E_0) * a * z * e / (k_B * T))
        axis = 0
    summands = (
        Δx * a * z * e / (k_B * T) * exponential / (1 + exponential)**2
        * (2 / (1 + exponential) - a * z * e / (k_B * T))
    )

    if individual:
        return [np.nan_to_num(s * (SOC_end - SOC_start)) for s in summands]
    else:
        return np.nan_to_num(np.sum(summands, axis=axis)) * (
            SOC_end - SOC_start
        )


def inverse_OCV_fit_function(SOC, *args, z=1.0, T=298.15, inverted=True):
    """!@brief The inverse of fitting_functions.OCV_fit_function.
    Basically OCV(SOC). Requires that the Δx entries are sorted by x.
    This corresponds to the parameters being sorted by decreasing E_0.

    @param E_OCV
        The SOCs for which the voltages shall be evaluated.
    @param args
        A list which length is dividable by three. These are the
        parameters E₀, a and Δx from "A parametric OCV model" in the
        order (E₀_0, a_0, Δx_0, E₀_1, a_1, Δx_1...).
    @param z
        The charge number of the electrode interface reaction.
    @param T
        The temperature of the electrode.
    @param inverted
        The default (False) uses the formulation from "A parametric
        OCV model". If True, the SOC argument gets flipped internally to
        give the more widely adopted convention for the SOC direction.
    @return
        The evaluation of the inverse model function.
    """

    if inverted:
        try:
            SOC = [1 - entry for entry in SOC]
        except TypeError:
            SOC = 1 - SOC

    E_0 = args[0::3]
    a = args[1::3]
    Δx = np.array(args[2::3])

    e = 1.602176634e-19
    k_B = 1.380649e-23

    # Define sensible bounds for the "root_scalar" optimizers.
    bracket = [np.min(E_0) - 7 * k_B * T / (np.min(np.abs(a)) * z * e),
               np.max(E_0) + 7 * k_B * T / (np.min(np.abs(a)) * z * e)]

    try:
        len(SOC)
    except TypeError:
        initial = E_0[(np.cumsum(Δx[::-1]) - SOC).argmin()]
        try:
            return root_scalar(
                lambda E: OCV_fit_function(E, *args, z=z, T=T)
                - SOC, method="toms748", bracket=bracket, x0=initial
            ).root
        #    fprime=lambda E: d_dE_OCV_fit_function(E, *args, z=z, T=T),
        #    fprime2=lambda E: d2_dE2_OCV_fit_function(E, *args, z=z, T=T)
        except ValueError:
            return 0.0

    roots = []
    try:
        initial = E_0[(np.cumsum(Δx[::-1]) - SOC[0]).argmin()]
    except ValueError:
        initial = E_0[0]
    for SOC_i in SOC:
        try:
            root = root_scalar(
                lambda E: OCV_fit_function(E, *args, z=z, T=T) - SOC_i,
                method="toms748", bracket=bracket, x0=initial
            ).root
            initial = root
        except ValueError:
            root = float("NaN")
        #    fprime=lambda E: d_dE_OCV_fit_function(E, *args, z=z, T=T),
        #    fprime2=lambda E: d2_dE2_OCV_fit_function(E, *args, z=z, T=T)
        roots.append(root)

    return np.array(roots)


def inverse_d_dSOC_OCV_fit_function(
    SOC, *args, z=1.0, T=298.15, inverted=True
):
    """!
    @brief The derivative of the inverse of fitting_functions.OCV_fit_function.
    Basically OCV'(SOC). Requires that the Δx entries are sorted by x.
    This corresponds to the parameters being sorted by decreasing E_0.

    @param E_OCV
        The SOCs for which the voltages shall be evaluated.
    @param args
        A list which length is dividable by three. These are the
        parameters E₀, a and Δx from "A parametric OCV model" in the
        order (E₀_0, a_0, Δx_0, E₀_1, a_1, Δx_1...).
    @param z
        The charge number of the electrode interface reaction.
    @param T
        The temperature of the electrode.
    @param inverted
        The default (False) uses the formulation from "A parametric
        OCV model". If True, the SOC argument gets flipped internally to
        give the more widely adopted convention for the SOC direction.
    @return
        The evaluation of the derivative of the inverse model function.
    """

    roots = np.array(inverse_OCV_fit_function(
        SOC, *args, z=z, T=T, inverted=inverted
    ))

    return 1 / d_dE_OCV_fit_function(roots, *args, z=z, T=T)


def inverse_d2_dSOC2_OCV_fit_function(
    SOC, *args, z=1.0, T=298.15, inverted=True
):
    """! The 2nd derivative of the inverse of .OCV_fit_function.
    Basically OCV'(SOC). Requires that the Δx entries are sorted by x.
    This corresponds to the parameters being sorted by decreasing E_0.

    @param E_OCV
        The SOCs for which the voltages shall be evaluated.
    @param args
        A list which length is dividable by three. These are the
        parameters E₀, a and Δx from "A parametric OCV model" in the
        order (E₀_0, a_0, Δx_0, E₀_1, a_1, Δx_1...).
    @param z
        The charge number of the electrode interface reaction.
    @param T
        The temperature of the electrode.
    @param inverted
        The default (False) uses the formulation from "A parametric
        OCV model". If True, the SOC argument gets flipped internally to
        give the more widely adopted convention for the SOC direction.
    @return
        The evaluation of the derivative of the inverse model function.
    """

    roots = np.array(inverse_OCV_fit_function(
        SOC, *args, z=z, T=T, inverted=inverted
    ))

    return (
        -d2_dE2_OCV_fit_function(roots, *args, z=z, T=T)
        / d_dE_OCV_fit_function(roots, *args, z=z, T=T)**3
    )


def fit_OCV(
    SOC,
    OCV,
    N=4,
    SOC_range_bounds=(0.2, 0.8),
    SOC_range_limits=(0.0, 1.0),
    z=1.0,
    T=298.15,
    inverted=True,
    fit_SOC_range=True,
    distance_order=2,
    weights=None,
    initial_parameters=None,
    minimize_options=None,
):
    """!@brief Fits data to fitting_functions.OCV_fit_function.
    In addition to the fit itself, a model-based correction
    to the provided SOC-OCV-data is made. If "SOC" lives in a
    (0,1)-range, the correction is given as its transformation to the
    (SOC_start, SOC_end)-range given as the first two returned numbers.
    If "other_electrode" is not None, the stoichiometric offset and the
    scaling that gives the (adjusted) SOC of this electrode given the
    SOC of the "other_electrode" are put before SOC_start and SOC_end.

    @param SOC
        The SOCs at which measurements were made.
    @param OCV
        The corresponding open-circuit voltages.
    @param N
        The number of phases of the OCV model.
    @param SOC_range_bounds
        Optional hard upper and lower bounds for the SOC correction from
        the left and the right side, respectively, as a 2-tuple. Use it
        as a limiting guess for the actual SOC range represented in the
        measurement. Has to be inside (0.0, 1.0). Set to (0.0, 1.0) to
        effectively disable SOC range estimation.
    @param SOC_range_limits
        Optional hard lower and upper bounds for the SOC correction from
        the left and the right side, respectively, as a 2-tuple. Use it
        if you know that your OCV data is incomplete and by how much.
        Has to be inside (0.0, 1.0). Set to (0.0, 1.0) to allow the
        SOC range estimation to assign datapoints to the asymptotes.
    @param z
        The charge number of the electrode interface reaction.
    @param T
        The temperature of the electrode.
    @param inverted
        If True (default), the widely adopted SOC convention is assumed.
        If False, the formulation of "A parametric OCV model" is used.
    @param fit_SOC_range
        If True (default), a model-based correction to the provided
        SOC-OCV-data is made.
    @param distance_order
        The order of the norm of the vector of the distances between OCV
        data and OCV model. Default is 2, i.e., the Euclidean norm.
        1 sets it to absolute distance, and float('inf') sets it to
        maximum distance. Note that 1 will lead to worse performance.
    @param weights
        Optional weights to apply to the vector of the distances between
        OCV data and OCV model. Defaults to equal weights.
    @param initial_parameters
        Optional initial guess for the model parameters. If left as-is,
        this will be automatically gleaned from the data. Use only if
        you have another fit to data of the same electrode material.
    @param minimize_options
        Dictionary that gets passed to scipy.optimize.minimize with the
        method 'trust-constr'. See scipy.optimize.show_options with the
        arguments 'minimize' and 'trust-constr' for details.
    @return
        The fitted parameters of fitting_functions.OCV_fit_function plus
        the fitted SOC range prepended.
    """

    e = 1.602176634e-19
    k_B = 1.380649e-23

    # The OCV fit function interprets SOC inversely to the popular way.
    if inverted:
        SOC = [1 - entry for entry in SOC]
    if weights is None:
        SOC, OCV = zip(*sorted(zip(SOC, OCV)))
        weights = np.array([1.0] * len(SOC))
    else:
        SOC, OCV, weights = zip(*sorted(zip(SOC, OCV, weights)))
        weights = np.array(weights)
    SOC = np.array(SOC)
    OCV = np.array(OCV)

    # Get the SOC correction bounds for the estimation procedure.
    SOC_start = SOC_range_bounds[0]
    SOC_end = SOC_range_bounds[1]
    SOC_limit_start = SOC_range_limits[0]
    SOC_limit_end = SOC_range_limits[1]
    if inverted:
        # [0, 1] ⊂ [-(y - 1), 1 - x]; x = x_start, y = x_end
        # [-(y - 1), 1 - x] → [0, 1], z ↦ (z + y - 1) / (y - x)
        # [0, 1] ↦ [(y - 1) / (y - x), y / (y - x)]
        x_start = (SOC_end - 1) / (SOC_end - SOC_start)
        x_end = SOC_end / (SOC_end - SOC_start)
        x_limit_start = (SOC_limit_end - 1) / (SOC_limit_end - SOC_limit_start)
        x_limit_end = SOC_limit_end / (SOC_limit_end - SOC_limit_start)
    else:
        # [0, 1] ⊂ [x, y]; x = x_start, y = x_end
        # [x, y] → [0, 1], z ↦ (z - x) / (y - x)
        # [0, 1] ↦ [- x / (y - x), (1 - x) / (y - x)]
        x_start = SOC_start / (SOC_start - SOC_end)
        x_end = (SOC_start - 1) / (SOC_start - SOC_end)
        x_limit_start = SOC_limit_start / (SOC_limit_start - SOC_limit_end)
        x_limit_end = (SOC_limit_start - 1) / (SOC_limit_start - SOC_limit_end)

    OCV_for_clustering = np.log(OCV.reshape(-1, 1))
    labels = KMeans(n_clusters=N, n_init='auto').fit_predict(
        OCV_for_clustering
    )
    E_0 = [np.median(OCV[labels == i]) for i in range(N)]
    Δx = []
    for i in range(N):
        Δx.append(np.abs(
            (SOC[labels == i][-1] - SOC[labels == i][0])
            / (SOC[-1] - SOC[0])
        ))
    # Δx = np.array([np.max([dx, 0.01]) for dx in Δx])
    # Δx = np.array(Δx) / np.sum(Δx)
    # Motivation for the weights: E(|X-µ|) = σ * √(2 / π)
    SOC_std_estimate = list(np.abs(
        SOC[1:-1] - 0.5 * (SOC[2:] + SOC[:-2])
    ) * np.sqrt(0.5 * np.pi))
    SOC_std_estimate = np.array(
        [SOC_std_estimate[0]]
        + SOC_std_estimate
        + [SOC_std_estimate[-1]]
    )
    ICA = smooth_fit(
        OCV, SOC, w=1 / SOC_std_estimate
    ).derivative()(OCV)
    slopes = [np.median(ICA[labels == i]) for i in range(N)]
    slopes = [s if s > 0.0 else 1.0 for s in slopes]

    if initial_parameters is None:
        initial = (
            [0.0, 1.0] * fit_SOC_range
            + [par for i in range(N)
               for par in (E_0[i], slopes[i], Δx[i])][:-1]
        )
    else:
        # Replace "a" with "slope" in the fit result.
        initial_a = initial_parameters[1::3]
        initial_Δx = initial_parameters[2::3]
        if len(initial_Δx) < len(initial_a):
            initial_Δx = np.append(initial_Δx, 1 - np.sum(initial_Δx))
        for i, (a, dx) in enumerate(zip(initial_a, initial_Δx)):
            initial_parameters[1 + 3 * i] = (
                -a / 4 * e / (k_B * T) * (dx * z)
            )
        initial = (
            [0.0, 1.0] * fit_SOC_range + list(initial_parameters)
        )

    E_0_ranges = [
        (np.min(OCV[labels == i]), np.max(OCV[labels == i])) for i in range(N)
    ]
    slope_ranges = [(0.0, 100.0) for _ in range(N)]
    Δx_ranges = [(0.0, 1.0) for _ in range(N)]

    bounds = list(zip(
        [x_start, x_limit_end] * fit_SOC_range
        + [par[0] for i in range(N)
           for par in (E_0_ranges[i], slope_ranges[i], Δx_ranges[i])][:-1],
        [x_limit_start, x_end] * fit_SOC_range
        + [par[1] for i in range(N)
           for par in (E_0_ranges[i], slope_ranges[i], Δx_ranges[i])][:-1]
    ))

    # Make sure that the sum of Δx without the implicit one of them is
    # not larger than 1.
    Δx_summation = np.array(
        [0, 0] * fit_SOC_range + (N - 1) * [0, 0, 1] + [0, 0]
    )
    Δx_constraint = so.LinearConstraint(Δx_summation, -np.inf, 1, True)

    if distance_order < float('inf'):
        if distance_order % 2:
            optimize_result = minimize(
                lambda x: np.sum(np.abs(
                    ((
                        OCV_fit_function(
                            OCV, *x, z=z, T=T, fit_SOC_range=True, rescale=True
                        ) - SOC
                    ) * weights)**distance_order / (
                        1 + d_dE_OCV_fit_function(
                            OCV, *x, z=z, T=T, fit_SOC_range=True, rescale=True
                        )**distance_order
                    )
                ))**(1 / distance_order),
                x0=initial,
                bounds=bounds,
                method='trust-constr',
                constraints=Δx_constraint,
                options=minimize_options,
            )
        else:
            optimize_result = minimize(
                lambda x: np.sum(
                    ((
                        OCV_fit_function(
                            OCV, *x, z=z, T=T, fit_SOC_range=True, rescale=True
                        ) - SOC
                    ) * weights)**distance_order / (
                        1 + d_dE_OCV_fit_function(
                            OCV, *x, z=z, T=T, fit_SOC_range=True, rescale=True
                        )**distance_order
                    )
                )**(1 / distance_order),
                jac='cs',
                x0=initial,
                bounds=bounds,
                method='trust-constr',
                constraints=Δx_constraint,
                options=minimize_options,
            )
    else:
        optimize_result = minimize(
            lambda x: np.amax(
                (OCV_fit_function(
                    OCV, *x, z=z, T=T, fit_SOC_range=True, rescale=True
                ) - SOC) * weights
            ),
            jac='cs',
            x0=initial,
            bounds=bounds,
            method='trust-constr',
            constraints=Δx_constraint,
            options=minimize_options,
        )
    fit = optimize_result.x
    slope_fit = fit[fit_SOC_range * 2 + 1::3]
    Δx_fit = fit[fit_SOC_range * 2 + 2::3]
    Δx_fit = np.append(Δx_fit, 1 - np.sum(Δx_fit))
    # Replace "slope" with "a" in the fit result.
    for i, (s, dx) in enumerate(zip(slope_fit, Δx_fit)):
        fit[fit_SOC_range * 2 + 1 + 3 * i] = (
            -s * 4 * (k_B / e) * T / (dx * z)
        )

    if fit_SOC_range:
        x_start = fit[0]
        x_end = fit[1]
        if inverted:
            # [0, 1] ⊂ [-(y - 1), 1 - x]; x = x_start, y = x_end
            # [-(y - 1), 1 - x] → [0, 1], z ↦ (z + y - 1) / (y - x)
            # [0, 1] ↦ [(y - 1) / (y - x), y / (y - x)]
            fit[0] = (x_end - 1) / (x_end - x_start)
            fit[1] = x_end / (x_end - x_start)
        else:
            # [0, 1] ⊂ [x, y]; x = x_start, y = x_end
            # [x, y] → [0, 1], z ↦ (z - x) / (y - x)
            # [0, 1] ↦ [- x / (y - x), (1 - x) / (y - x)]
            fit[0] = - x_start / (x_end - x_start)
            fit[1] = (1 - x_start) / (x_end - x_start)
    else:
        fit = [0, 1] + fit

    return OCV_fit_result(fit, SOC, OCV, optimize_result=optimize_result)


def verbose_spline_parameterization(
    coeffs,
    knots,
    order,
    format='python',
    function_name="OCV",
    function_args="SOC",
    derivatives=0,
    spline_transformation='',
    verbose=False,
):
    """!@brief Gives the monomic representation of a B-spline.
    @param coeffs
        The B-spline coefficients as structured by scipy.interpolate.
    @param knots
        The B-spline knots as structured by scipy.interpolate.
    @param order
        The order of the B-spline.
    @param format
        Gives the file/language format for the function representation.
        Default is 'python'. The other choice is 'matlab'.
    @param function_name
        An optional name for the printed Python function.
    @param function_args
        An optional string for the arguments of the function.
    @param derivatives
        The derivatives of the spline to also include in the return.
        Default is 0, which gives the spline. 1 would give the spline,
        followed by its derivative. Can not be higher than spline order.
        Derivatives are only continuous when derivatives < order.
    @param spline_transformation
        Give a string if you want to include a function that gets
        applied to the whole spline, e.g. 'exp'.
    @param verbose
        Print information about the progress of the conversion.
    @return
        A string that gives a Python function when "exec"-uted.
    """

    if derivatives > 1:
        raise NotImplementedError(
            "Spline derivatives greater than 1 are not implemented."
        )
    if derivatives > order:
        raise ValueError(
            "A " + str(order) + "-order spline can not be differentiated "
            + str(derivatives) + " times."
        )

    # Ensure that 'function_name' and 'function_args' are valid Python
    # variable names. This convention should work for MatLab as well.
    function_name = re.sub(r'\W|^(?=\d)', '_', function_name)
    function_args = re.sub(r'\W|^(?=\d)', '_', function_args)

    tab = " " * 4
    nl = " ...\r\n" if format == 'matlab' else "\r\n"
    if format == 'python':
        func = "def " + function_name + "(" + function_args + "):" + nl
        func += tab + "return " + spline_transformation + "("
    elif format == 'matlab':
        func = ("function [U] = " + function_name + "(" + function_args
                + ")\r\nU = (")
    else:
        func = ""

    t = [knots[0]] * order + list(knots) + [knots[-1]] * order

    if order == 1:

        # "p_c": shorthand for piecewise coefficients.
        p_c = np.zeros([len(knots) - 1, order + 1])

        for i in range(len(knots) - 1 + order):
            if t[i] != t[i + 1]:
                p_c[i - order][1] = p_c[i - order][1] + coeffs[i] * (
                    1 / (t[i + 1] - t[i])
                )
                p_c[i - order][0] = p_c[i - order][0] + coeffs[i] * (
                    -t[i] / (t[i + 1] - t[i])
                )
            if t[i + 1] != t[i + 2]:
                p_c[i - order + 1][1] = p_c[i + 1 - order][1] + coeffs[i] * (
                    -1 / (t[i + 2] - t[i + 1])
                )
                p_c[i - order + 1][0] = p_c[i + 1 - order][0] + coeffs[i] * (
                    t[i + 2] / (t[i + 2] - t[i + 1])
                )

        for i, poly in enumerate(p_c):
            func += nl
            if i == 0:
                if len(p_c) > 1:
                    func += (tab * 2 + "  (" + function_args + " < "
                             + str(knots[i + 1]) + ") * ")
                else:
                    func += tab * 2 + "  "
            elif i == len(p_c) - 1:
                func += (tab * 2 + "+ (" + function_args + " >= "
                         + str(knots[i]) + ") * ")
            else:
                func += (tab * 2 + "+ (" + function_args + " >= "
                         + str(knots[i]) + ") * (" + function_args + " < "
                         + str(knots[i + 1]) + ") * ")
            func += ("(" * (len(poly) - 1) + nl + tab * 3 + "  "
                     + str(poly[order]))
            for j in range(len(poly) - 2):
                func += (" * " + function_args + "" + nl + tab * 3 + "+ "
                         + str(poly[order - j - 1]) + ")")
            func += (" * " + function_args + nl + tab * 3 + "+ " + str(poly[0])
                     + ")")

        func += (nl + tab + ")\r\nend" if format == 'matlab' else
                 nl + tab + ")" + nl)

        if derivatives == 1:

            func += nl
            if format == 'python':
                func += ("def derivative_" + function_name + "("
                         + function_args + "):" + nl)
                func += tab + "return " + spline_transformation + "("
            elif format == 'matlab':
                func += ("function [U] = derivative_" + function_name + "("
                         + function_args + ")\r\nU = (")

            for i, poly in enumerate(p_c):
                func += nl
                if i == 0:
                    if len(p_c) > 1:
                        func += (tab * 2 + "  (" + function_args + " < "
                                 + str(knots[i + 1]) + ") * ")
                    else:
                        func += tab * 2 + "  "
                elif i == len(p_c) - 1:
                    func += (tab * 2 + "+ (" + function_args + " >= "
                             + str(knots[i]) + ") * ")
                else:
                    func += (tab * 2 + "+ (" + function_args + " >= "
                             + str(knots[i]) + ") * (" + function_args + " < "
                             + str(knots[i + 1]) + ") * ")
                # func += ("(" * (len(poly) - 2) + nl + tab * 3 + "  "
                #          + str(order * poly[order]))
                # for j in range(len(poly) - 3):
                #     func += (" * " + function_args + "" + nl + tab * 3 + "+ "
                #              + str((order - j - 1) * poly[order - j - 1])
                #              + ")")
                func += "(" + nl + tab * 3 + str(poly[1]) + ")"

            func += (nl + tab + ")\r\nend" if format == 'matlab' else
                     nl + tab + ")" + nl)

    elif order == 2:

        # "p_c": shorthand for piecewise coefficients.
        p_c = np.zeros([len(knots) - 1, order + 1])

        for i in range(len(knots) - 1 + order):
            if t[i] != t[i + 1] and t[i] != t[i + 2]:
                p_c[i - order][2] = p_c[i - order][2] + coeffs[i] * (
                    1 / ((t[i + 2] - t[i]) * (t[i + 1] - t[i]))
                )
                p_c[i - order][1] = p_c[i - order][1] + coeffs[i] * (
                    -2 * t[i] / ((t[i + 2] - t[i]) * (t[i + 1]-t[i]))
                )
                p_c[i - order][0] = p_c[i - order][0] + coeffs[i] * (
                    t[i]**2 / ((t[i + 2] - t[i]) * (t[i + 1]-t[i]))
                )
            if t[i + 1] != t[i + 2]:
                if t[i] != t[i + 2]:
                    p_c[i - order + 1][2] = (
                        p_c[i - order + 1][2] + coeffs[i] * (
                            -1 / ((t[i + 2] - t[i + 1]) * (t[i + 2] - t[i]))
                        )
                    )
                    p_c[i - order + 1][1] = (
                        p_c[i - order + 1][1] + coeffs[i] * (
                            (t[i + 2] + t[i])
                            / ((t[i + 2] - t[i + 1]) * (t[i + 2] - t[i]))
                        )
                    )
                    p_c[i - order + 1][0] = (
                        p_c[i - order + 1][0] + coeffs[i] * (
                            -t[i]
                            * t[i + 2] / ((t[i + 2] - t[i + 1])
                                          * (t[i + 2] - t[i]))
                        )
                    )
                if t[i + 1] != t[i + 3]:
                    p_c[i - order + 1][2] = p_c[i - order + 1][2] + coeffs[i]*(
                        -1 / ((t[i + 2] - t[i + 1]) * (t[i + 3] - t[i + 1])))
                    p_c[i - order + 1][1] = p_c[i - order + 1][1] + coeffs[i]*(
                        (t[i + 3] + t[i + 1]) / ((t[i + 2] - t[i + 1])
                                                 * (t[i + 3] - t[i + 1])))
                    p_c[i - order + 1][0] = p_c[i - order + 1][0] + coeffs[i]*(
                        -t[i + 3] * t[i + 1] / ((t[i + 2] - t[i + 1])
                                                * (t[i + 3] - t[i + 1])))
            if t[i + 1] != t[i + 3] and t[i + 2] != t[i + 3]:
                p_c[i - order + 2][2] = p_c[i - order + 2][2] + coeffs[i] * (
                    1 / ((t[i + 3] - t[i + 1]) * (t[i + 3] - t[i + 2]))
                )
                p_c[i - order + 2][1] = p_c[i - order + 2][1] + coeffs[i] * (
                    -2 * t[i + 3]
                    / ((t[i + 3] - t[i + 1]) * (t[i + 3] - t[i + 2]))
                )
                p_c[i - order + 2][0] = p_c[i - order + 2][0] + coeffs[i] * (
                    t[i + 3]**2
                    / ((t[i + 3] - t[i + 1]) * (t[i + 3] - t[i + 2]))
                )

        for i, poly in enumerate(p_c):
            func += nl
            if i == 0:
                if len(p_c) > 1:
                    func += (tab * 2 + "  (" + function_args + " < "
                             + str(knots[i + 1]) + ") * ")
                else:
                    func += tab * 2 + "  "
            elif i == len(p_c) - 1:
                func += (tab * 2 + "+ (" + function_args + " >= "
                         + str(knots[i]) + ") * ")
            else:
                func += (tab * 2 + "+ (" + function_args + " >= "
                         + str(knots[i]) + ") * (" + function_args + " < "
                         + str(knots[i + 1]) + ") * ")
            func += ("(" * (len(poly) - 1) + nl + tab * 3 + "  "
                     + str(poly[order]))
            for j in range(len(poly) - 2):
                func += (" * " + function_args + nl + tab * 3 + "+ "
                         + str(poly[order - j - 1]) + ")")
            func += (" * " + function_args + nl + tab * 3 + "+ " + str(poly[0])
                     + ")")

        func += (nl + tab + ")\r\nend" if format == 'matlab' else
                 nl + tab + ")" + nl)

        if derivatives == 1:

            func += nl
            if format == 'python':
                func += ("def derivative_" + function_name + "("
                         + function_args + "):" + nl)
                func += tab + "return " + spline_transformation + "("
            elif format == 'matlab':
                func += ("function [U] = derivative_" + function_name + "("
                         + function_args + ")\r\nU = (")

            for i, poly in enumerate(p_c):
                func += nl
                if i == 0:
                    if len(p_c) > 1:
                        func += (tab * 2 + "  (" + function_args + " < "
                                 + str(knots[i + 1]) + ") * ")
                    else:
                        func += tab * 2 + "  "
                elif i == len(p_c) - 1:
                    func += (tab * 2 + "+ (" + function_args + " >= "
                             + str(knots[i]) + ") * ")
                else:
                    func += (tab * 2 + "+ (" + function_args + " >= "
                             + str(knots[i]) + ") * (" + function_args + " < "
                             + str(knots[i + 1]) + ") * ")
                func += ("(" * (len(poly) - 2) + nl + tab * 3 + "  "
                         + str(order * poly[order]))
                for j in range(len(poly) - 3):
                    func += (" * " + function_args + nl + tab * 3 + "+ "
                             + str((order - j - 1) * poly[order - j - 1])
                             + ")")
                func += (" * " + function_args + nl + tab * 3 + "+ "
                         + str(poly[1]) + ")")

            func += (nl + tab + ")\r\nend" if format == 'matlab' else
                     nl + tab + ")" + nl)

    else:

        if len(knots) > 13 - 2 * order:
            raise ValueError(
                "Symbolical computation of a higher-order spline "
                "with this many knots would take too long. Reduce the amount "
                "of knots or use an order 2 spline for dense interpolation."
            )

        x = sp.Symbol("x")

        # Definition of B-splines by recursion.
        def B(i, k=order):

            if k > 0:
                if t[i + k] == t[i] and t[i + k + 1] == t[i + 1]:
                    return 0
                elif t[i + k] == t[i]:
                    return ((t[i + k + 1] - x) / (t[i + k + 1] - t[i + 1])
                            * B(i + 1, k - 1))
                elif t[i + k + 1] == t[i + 1]:
                    return (x - t[i]) / (t[i + k] - t[i]) * B(i, k - 1)
                else:
                    return ((x - t[i]) / (t[i + k] - t[i]) * B(i, k - 1)
                            + (t[i + k + 1] - x) / (t[i + k + 1] - t[i + 1])
                            * B(i + 1, k - 1))

            else:
                # Note: "&" is the bit-wise operator for integers and the
                # "sp.And(⋅,⋅)" function for everything else for convenience.
                # Another note: later in the code, it is asssumed that "x"
                # is on the lhs and the interval boundary on the rhs.
                return sp.Piecewise(
                    (1, sp.And(x >= t[i], x < t[i+1])),
                    (0, True)
                )

        def B_derivative(i, k=order):

            if k > 0:
                if t[i + k] == t[i] and t[i + k + 1] == t[i + 1]:
                    return 0
                elif t[i + k] == t[i]:
                    return -k / (t[i + k + 1] - t[i + 1]) * B(i + 1, k - 1)
                elif [i + k + 1] == t[i + 1]:
                    return k / (t[i + k] - t[i]) * B(i, k - 1)
                else:
                    return (
                        k / (t[i + k] - t[i]) * B(i, k - 1)
                        - k / (t[i + k + 1] - t[i + 1])
                        * B(i + 1, k - 1)
                    )

            else:
                return 0

        if verbose:
            sp.init_printing()
            print("Simplifying the B-splines. This may take a while...")

        B_splines = []
        for i in range(len(knots) - 1 + order):
            print("Simplifying B-spline #" + str(i+1) + " of "
                  + str(len(knots) - 1 + order))
            B_splines.append(sp.simplify(B(i, order)))

        if derivatives == 1:
            B_spline_derivatives = []
            print("Simplifying derivated B-spline #" + str(i+1) + " of "
                  + str(len(knots) - 1 + order))
            B_spline_derivatives.append(sp.simplify(B_derivative(i, order)))

        if verbose:
            print("Simplifying the spline. This may take another while...")
        ip_spline = 0
        for i, B_spline in enumerate(B_splines):
            ip_spline = ip_spline + coeffs[i] * B_spline
            ip_spline = sp.simplify(ip_spline)
        if derivatives == 1:
            ip_spline_der = 0
            for i, B_spline_derivative in enumerate(B_spline_derivatives):
                ip_spline_der = ip_spline_der + coeffs[i] * B_spline_derivative
                ip_spline_der = sp.simplify(ip_spline_der)
        if verbose:
            print("Interpolating spline:")
            print()
            sp.pprint(ip_spline)
            print()
            if derivatives == 1:
                print()
                sp.pprint(ip_spline_der)
                print()

        if derivatives == 0:
            polys = [ip_spline]
        elif derivatives == 1:
            polys = [ip_spline, ip_spline_der]

        for i, poly in enumerate(polys):

            if i == 1:
                func += nl
                if format == 'python':
                    func += ("def derivative_" + function_name + "("
                             + function_args + "):" + nl)
                    func += tab + "return " + spline_transformation + "("
                elif format == 'matlab':
                    func += ("function [U] = derivative_" + function_name + "("
                             + function_args + ")\r\nU = (")

            polys_sorted_by_definition = [[], [], []]
            for part in ip_spline.args:
                try:
                    polynomial = sp.Poly(part[0])
                except sp.polys.polyerrors.GeneratorsNeeded:
                    # "part[0]" was 0 probably.
                    continue
                polys_sorted_by_definition[0].append(part[1].args[0].rhs)
                polys_sorted_by_definition[1].append(part[1].args[1].rhs)
                polys_sorted_by_definition[2].append(polynomial.all_coeffs())
            polys_sorted_by_definition = list(sorted(zip(
                *polys_sorted_by_definition)))

            for i, (left, right, poly) in enumerate(
                polys_sorted_by_definition
            ):
                func += nl
                if i == 0:
                    if len(polys_sorted_by_definition) > 1:
                        func += (tab * 2 + "  (" + function_args + " < "
                                 + str(right) + ") * ")
                    else:
                        func += tab * 2 + "  "
                elif i == len(polys_sorted_by_definition) - 1:
                    func += (tab * 2 + "+ (" + function_args + " >= "
                             + str(left) + ") * ")
                else:
                    func += (tab * 2 + "+ (" + function_args + " >= "
                             + str(left) + ") * (" + function_args + " < "
                             + str(right) + ") * ")
                func += ("(" * (len(poly) - 1) + nl + tab * 3 + "  "
                         + str(poly[0]))
                for j in range(len(poly) - 2):
                    func += (" * " + function_args + nl + tab * 3 + "+ "
                             + str(poly[j + 1]) + ")")
                func += (" * " + function_args + nl + tab * 3 + "+ "
                         + str(poly[-1]) + ")")
            func += (nl + tab + ")\r\nend" if format == 'matlab' else
                     nl + tab + ")" + nl + nl)

    return func



"""!@package ep_bolfi.utility.dataset_formatting
Defines datatypes for further processing.
"""

import csv
import h5py
import json
import re

from ep_bolfi.utility.fitting_functions import fit_exponential_decay
from math import pi
from os import linesep


class Measurement:
    """!@brief Defines common methods for measurement objects. """

    def to_json(self):
        raise NotImplementedError

    @classmethod
    def from_json(cls, json_string):
        raise NotImplementedError

    def table_descriptors(self):
        raise NotImplementedError

    @classmethod
    def table_mapping(cls):
        raise NotImplementedError

    def segment_tables(self, start=0, stop=None, step=1):
        raise NotImplementedError

    def example_table_row(self):
        raise NotImplementedError

    def subslice(self, start, stop, step=1):
        raise NotImplementedError

    def subarray(self, array):
        raise NotImplementedError

    def extend(self, other):
        raise NotImplementedError


class Cycling_Information(Measurement):
    """!@brief Contains basic cycling informations.
    Each member variable is a list and has the same length as the other ones.
    """

    def __init__(
        self, timepoints, currents, voltages, other_columns={}, indices=None
    ):
        ## The times at which measurements were taken. Usually a list of
        # lists where each list corresponds to a segment of the
        # measurement.
        self.timepoints = timepoints
        ## The measured current at those times. A list which usually
        # contains lists for segments with variable current and floats
        # for segments with constant current.
        self.currents = currents
        ## The measured voltage at those times. A list which usually
        # contains lists for segments with variable voltage and floats
        # for segments with constant voltage.
        self.voltages = voltages
        ## The contents of any other columns. A dictionary ("columns")
        # which values are lists which contain lists for segments.
        # The keys should match user input for the columns.
        self.other_columns = other_columns
        if indices is None:
            ## The indices of the individual segments. Defaults to a simple
            # numbering of the segments present. May be used for plotting
            # purposes, e.g., for colourcoding the segments by cycle.
            self.indices = [i for i in range(len(self))]
        else:
            self.indices = indices
        # This syntax doesn't work when arrays are involved, apparently.
        # self.indices = indices or [i for i in range(len(self))]

    def __len__(self):
        return max([
            len(self.timepoints),
            len(self.currents),
            len(self.voltages),
            *[len(column) for column in self.other_columns.values()]
        ])

    def to_json(self):
        return json.dumps({
            "timepoints [s]": self.timepoints,
            "currents [A]": self.currents,
            "voltages [V]": self.voltages,
            "other_columns": self.other_columns,
            "indices": self.indices,
        })

    @classmethod
    def from_json(cls, json_string):
        return cls(*(json.loads(json_string).values()))

    def table_descriptors(self):
        descriptors = ["indices"]
        if self.timepoints:
            descriptors.append("timepoints [s]")
        if self.currents:
            descriptors.append("currents [A]")
        if self.voltages:
            descriptors.append("voltages [V]")
        descriptors.extend([name for name in self.other_columns.keys()])
        return descriptors

    @classmethod
    def table_mapping(cls):
        return {
            'indices': "indices",
            'timepoints': "timepoints [s]",
            'currents': "currents [A]",
            'voltages': "voltages [V]",
        }

    def segment_tables(self, start=0, stop=None, step=1):
        if stop is None:
            stop = len(self)
        for i in range(start, stop, step):
            segment = [self.indices[i]]
            if self.timepoints:
                segment.append(self.timepoints[i])
            if self.currents:
                segment.append(self.currents[i])
            if self.voltages:
                segment.append(self.voltages[i])
            segment.extend([
                column[i] for column in self.other_columns.values()
            ])
            yield segment

    def example_table_row(self):
        try:
            example = []
            example.append([self.indices[0][0]])
            if self.timepoints:
                example.append([self.timepoints[0][0]])
            if self.currents:
                example.append([self.currents[0][0]])
            if self.voltages:
                example.append([self.voltages[0][0]])
            example.extend([
                [column[0][0]]
                for column in self.other_columns.values()
            ])
            return example
        except (TypeError, IndexError):
            print("Transforming indices into columns for tabular view.")
            lengths_of_segments = []
            for i in range(len(self)):
                try:
                    length = len(self.timepoints[i])
                    if length > 0:
                        lengths_of_segments.append(length)
                        continue
                except IndexError:
                    pass
                key = list(self.other_columns.keys())[0]
                lengths_of_segments.append((len(self.other_columns[key][i])))
            self.indices = [
                [index] * length
                for index, length in zip(self.indices, lengths_of_segments)
            ]
            example = []
            example.append([self.indices[0][0]])
            if self.timepoints:
                example.append([self.timepoints[0][0]])
            if self.currents:
                example.append([self.currents[0][0]])
            if self.voltages:
                example.append([self.voltages[0][0]])
            example.extend([
                [column[0][0]]
                for column in self.other_columns.values()
            ])
            return example

    def subslice(self, start, stop, step=1):
        """!@brief Selects a subslice of the data segments.
        The arguments exactly match the slice(...) notation.
        @param start
            The index of the first segment to be included.
        @param stop
            The index of the first segment to not be included.
        @param step
            Steps between selected segments. Default: 1.
        @return
            A new Cycling_Information object containing the slice.
        """

        return Cycling_Information(
            self.timepoints[start:stop:step],
            self.currents[start:stop:step],
            self.voltages[start:stop:step],
            {name: column[start:stop:step]
             for name, column in self.other_columns.items()},
            self.indices[start:stop:step]
        )

    def subarray(self, array):
        """!@brief Selects the data segments with the given indices.
        @param array
            The indices of the segments that are to be returned.
        @return
            A new Cycling_Information object containing the subset.
        """

        return Cycling_Information(
            [self.timepoints[index] for index in array],
            [self.currents[index] for index in array],
            [self.voltages[index] for index in array],
            {name: [column[index] for index in array]
             for name, column in self.other_columns.items()},
            [self.indices[index] for index in array]
        )

    def extend(self, other):
        other_column_names = set(self.other_columns.keys()).intersection(
            set(other.other_columns.keys())
        )
        self.timepoints.extend(other.timepoints)
        self.currents.extend(other.currents)
        self.voltages.extend(other.voltages)
        for name in other_column_names:
            self.other_columns[name].extend(other.other_columns[name])
        self.indices.extend(other.indices)


class Static_Information(Cycling_Information):
    """!@brief Contains additional informations, e.g. for GITT.
    Each member variable is a list and has the same length as the other
    ones.
    """

    def __init__(
        self,
        timepoints,
        currents,
        voltages,
        other_columns={},
        indices=None
    ):
        super().__init__(timepoints, currents, voltages, other_columns,
                         indices)
        asymptotic_voltages = [-1.0] * len(timepoints)
        ir_steps = [0.0] * len(timepoints)
        last_U = 0
        exp_U_decays = []
        exp_I_decays = []

        for i, (t, I, U) in enumerate(zip(
                timepoints, currents, voltages)):
            for curve in fit_exponential_decay(t, I):
                I_0, ΔI, I_decay = curve[2]
                exp_I_decays.append((I_0, ΔI, I_decay))
            for j, curve in enumerate(fit_exponential_decay(t, U)):
                U_0, ΔU, U_decay = curve[2]
                exp_U_decays.append((U_0, ΔU, U_decay))
                if j == 0:
                    asymptotic_voltages[i] = U_0
            if asymptotic_voltages[i] == -1.0:
                asymptotic_voltages[i] = U[-1]
            ir_steps[i] = U[0] - last_U
            last_U = U[-1]
        ir_steps[0] = 0.0

        ## The voltages that the voltage curve seems to converge to in a
        # segment. Only makes sense for those segments that are rest
        # periods or when the OCV was subtracted.
        self.asymptotic_voltages = asymptotic_voltages
        ## The instantaneous IR drops before each segment. Positive
        # values are voltage rises and negative values voltage drops.
        self.ir_steps = ir_steps
        ## Same as exp_U_decays for current decays (PITT).
        #
        self.exp_I_decays = exp_I_decays
        ## The fit parameters of the exponential voltage decays in each
        # segment. Each set of fit parameters is a 3-tuple (a,b,c) where
        # the fit function has the following form:
        #     a + b * exp(-c * (t - t_end_of_segment)).
        # Failed or missing fits are best indicated by (NaN, NaN, NaN). """
        self.exp_U_decays = exp_U_decays

    def to_json(self):
        return json.dumps({
            "timepoints [s]": self.timepoints,
            "currents [A]": self.currents,
            "voltages [V]": self.voltages,
            "other_columns": self.other_columns,
            "indices": self.indices,
            "asymptotic voltages [V]": self.asymptotic_voltages,
            "IR steps [V]": self.ir_steps,
            "exponential current decays (fit parameters)": self.exp_I_decays,
            "exponential voltage decays (fit parameters)": self.exp_U_decays
        })

    def subslice(self, start, stop, step=1):
        """!@brief Selects a subslice of the data segments.
        The arguments exactly match the slice(...) notation.
        @param start
            The index of the first segment to be included.
        @param stop
            The index of the first segment to not be included.
        @param step
            Steps between selected segments. Default: 1.
        @return
            A new Static_Information object containing the slice.
        """

        return Static_Information(
            self.timepoints[start:stop:step],
            self.currents[start:stop:step],
            self.voltages[start:stop:step],
            self.asymptotic_voltages[start:stop:step],
            self.ir_steps[start:stop:step],
            self.exp_I_decays[start:stop:step],
            self.exp_U_decays[start:stop:step],
            {name: column[start:stop:step]
             for name, column in self.other_columns.items()},
            self.indices[start:stop:step]
        )

    def subarray(self, array):
        """!@brief Selects the data segments with the given indices.
        @param array
            The indices of the segments that are to be returned.
        @return
            A new Static_Information object containing the subset.
        """

        return Static_Information(
            [self.timepoints[index] for index in array],
            [self.currents[index] for index in array],
            [self.voltages[index] for index in array],
            [self.asymptotic_voltages[index] for index in array],
            [self.ir_steps[index] for index in array],
            [self.exp_I_decays[index] for index in array],
            [self.exp_U_decays[index] for index in array],
            {name: [column[index] for index in array]
             for name, column in self.other_columns.items()},
            [self.indices[index] for index in array]
        )

    def extend(self, other):
        other_column_names = set(self.other_columns.keys()).intersection(
            set(other.other_columns.keys())
        )
        self.timepoints.extend(other.timepoints)
        self.currents.extend(other.currents)
        self.voltages.extend(other.voltages)
        self.asymptotic_voltages.extend(other.asymptotic_voltages)
        self.ir_steps.extend(other.ir_steps)
        self.exp_I_decays.extend(other.exp_I_decays)
        self.exp_U_decays.extend(other.exp_U_decays)
        for name in other_column_names:
            self.other_columns[name].extend(other.other_columns[name])
        self.indices.extend(other.indices)


class Impedance_Measurement(Measurement):
    """!@brief Contains basic impedance data.
    Each member variable is a list and has the same length as the other
    ones.
    """

    def __init__(
        self,
        frequencies,
        real_impedances,
        imaginary_impedances,
        phases,
        other_columns={},
        indices=None
    ):
        ## The frequencies at which impedances were measured. Usually a list
        # of lists where each list corresponds to a different equilibrium.
        self.frequencies = frequencies
        ## The real part of the impedances measured at those frequencies.
        self.real_impedances = real_impedances
        ## The imaginary part of the impedances measured at those
        # frequencies.
        self.imaginary_impedances = imaginary_impedances
        ## The phases of the impedance measured at those frequencies.
        self.phases = phases
        ## The contents of any other columns. A dictionary ("columns")
        # which values are lists which contain lists for segments.
        # The keys should match user input for the columns.
        self.other_columns = other_columns
        ## The indices of the individual segments. Defaults to a simple
        # numbering of the segments present. May be used for plotting
        # purposes, e.g., for colourcoding the segments by cycle.
        if indices is None:
            self.indices = [i for i in range(len(self))]
        else:
            self.indices = indices

    def __len__(self):
        return max([
            len(self.frequencies),
            len(self.real_impedances),
            len(self.imaginary_impedances),
            len(self.phases),
            *[len(column) for column in self.other_columns.values()]
        ])

    @property
    def complex_impedances(self):
        return [
            [r + 1j * i for r, i in zip(real_segment, imag_segment)]
            for real_segment, imag_segment
            in zip(self.real_impedances, self.imaginary_impedances)
        ]

    def to_json(self):
        return json.dumps({
            "frequencies [Hz]": self.frequencies,
            "real_impedances [Ω]": self.real_impedances,
            "imaginary_impedances [Ω]": self.imaginary_impedances,
            "phases [rad]": self.phases,
            "other_columns": self.other_columns,
            "indices": self.indices,
        })

    @classmethod
    def from_json(cls, json_string):
        return cls(*(json.loads(json_string).values()))

    def table_descriptors(self):
        descriptors = ["indices"]
        if self.frequencies:
            descriptors.append("frequencies [Hz]")
        if self.real_impedances:
            descriptors.append("real_impedances [Ω]")
        if self.imaginary_impedances:
            descriptors.append("imaginary_impedances [Ω]")
        if self.phases:
            descriptors.append("phases [rad]")
        descriptors.extend([name for name in self.other_columns.keys()])
        return descriptors

    @classmethod
    def table_mapping(cls):
        return {
            'indices': "indices",
            'frequencies': "frequencies [Hz]",
            'real_impedances': "real_impedances [Ω]",
            'imaginary_impedances': "imaginary_impedances [Ω]",
            'phases': "phases [rad]",
        }

    def segment_tables(self, start=0, stop=None, step=1):
        if stop is None:
            stop = len(self)
        for i in range(start, stop, step):
            segment = [self.indices[i]]
            if self.frequencies:
                segment.append(self.frequencies[i])
            if self.real_impedances:
                segment.append(self.real_impedances[i])
            if self.imaginary_impedances:
                segment.append(self.imaginary_impedances[i])
            if self.phases:
                segment.append(self.phases[i])
            segment.extend([
                column[i] for column in self.other_columns.values()
            ])
            yield segment

    def example_table_row(self):
        try:
            example = []
            example.append([self.indices[0][0]])
            if self.frequencies:
                example.append([self.frequencies[0][0]])
            if self.real_impedances:
                example.append([self.real_impedances[0][0]])
            if self.imaginary_impedances:
                example.append([self.imaginary_impedances[0][0]])
            if self.phases:
                example.append([self.phases[0][0]])
            example.extend([
                [column[0][0]]
                for column in self.other_columns.values()
            ])
            return example
        except (TypeError, IndexError):
            print("Transforming indices into columns for tabular view.")
            lengths_of_segments = []
            for i in range(len(self)):
                try:
                    length = len(self.frequencies[i])
                    if length > 0:
                        lengths_of_segments.append(length)
                        continue
                except IndexError:
                    pass
                key = list(self.other_columns.keys())[0]
                lengths_of_segments.append((len(self.other_columns[key][i])))
            self.indices = [
                [index] * length
                for index, length in zip(self.indices, lengths_of_segments)
            ]
            example = []
            example.append([self.indices[0][0]])
            if self.frequencies:
                example.append([self.frequencies[0][0]])
            if self.real_impedances:
                example.append([self.real_impedances[0][0]])
            if self.imaginary_impedances:
                example.append([self.imaginary_impedances[0][0]])
            if self.phases:
                example.append([self.phases[0][0]])
            example.extend([
                [column[0][0]]
                for column in self.other_columns.values()
            ])
            return example

    def subslice(self, start, stop, step=1):
        """!@brief Selects a subslice of the data segments.
        The arguments exactly match the slice(...) notation.
        @param start
            The index of the first segment to be included.
        @param stop
            The index of the first segment to not be included.
        @param step
            Steps between selected segments. Default: 1.
        @return
            A new Impedance_Measurement object containing the slice.
        """

        return Impedance_Measurement(
            self.frequencies[start:stop:step],
            self.real_impedances[start:stop:step],
            self.imaginary_impedances[start:stop:step],
            self.phases[start:stop:step],
            {name: column[start:stop:step]
             for name, column in self.other_columns.items()},
            self.indices[start:stop:step]
        )

    def subarray(self, array):
        """!@brief Selects the data segments with the given indices.
        @param array
            The indices of the segments that are to be returned.
        @return
            A new Impedance_Measurement object containing the subset.
        """

        return Impedance_Measurement(
            [self.frequencies[index] for index in array],
            [self.real_impedances[index] for index in array],
            [self.imaginary_impedances[index] for index in array],
            [self.phases[index] for index in array],
            {name: [column[index] for index in array]
             for name, column in self.other_columns.items()},
            [self.indices[index] for index in array]
        )

    def extend(self, other):
        other_column_names = set(self.other_columns.keys()).intersection(
            set(other.other_columns.keys())
        )
        self.frequencies.extend(other.frequencies)
        self.real_impedances.extend(other.real_impedances)
        self.imaginary_impedances.extend(other.imaginary_impedances)
        self.phases.extend(other.phases)
        for name in other_column_names:
            self.other_columns[name].extend(other.other_columns[name])
        self.indices.extend(other.indices)


def read_csv_from_measurement_system(
    path,
    encoding,
    number_of_comment_lines,
    headers,
    delimiter='\t',
    decimal='.',
    datatype="cycling",
    segment_column=-1,
    segments_to_process=None,
    current_sign_correction={},
    correction_column=-1,
    flip_voltage_sign=False,
    flip_imaginary_impedance_sign=False,
    max_number_of_lines=-1
):
    """!@brief Read the measurements as returned by common instruments.
    Example: cycling measurements from Basytec devices. Their format
    resembles a csv file with one title and one header comment line.
    So the first line will be ignored and the second used for headers.
    @param path
        The full or relative path to the measurement file.
    @param encoding
        The encoding of that file, e.g. "iso-8859-1".
    @param number_of_comment_lines
        The number of lines that have to be skipped over in order to
        arrive at the first dataset line.
    @param headers
        A dictionary. Its keys are the indices of the columns
        which are to be read in. The corresponding values are there to
        tell this function which kind of data is in which column. The
        following format has to be used: "<name> [<unit>]" where "name"
        is "U" (voltage), "I" (current) or "t" (time) and "unit" is
        "V", "A", "h", "m" or "s" with the optional prefixes "k", "m",
        "µ" or "n". This converts the data to prefix-less SI units.
        Additional columns may be read in with keys not in this format.
        The columns for segments and sign correction are only given by
        #segment_column and #correction_column.
    @param delimiter
        The delimiter string between datapoints. The default is "\t".
    @param decimal
        The string used for the decimal point. Default: ".".
    @param datatype
        Default is "cycling", where cycling information is assumed in
        the file. "static" will trigger the additional extraction of
        exponential decays that are relevant to e.g. GITT.
        "impedance" will treat the file as an impedance measurement
        with frequencies and impedances instead of time and voltage.
    @param segment_column
        The index of the column that stores the index
        of the current segment. If it changes from one data point to the
        next, that is used as the dividing line between two segments.
        Default is -1, which returns the dataset in one segment.
    @param segments_to_process
        A list of indices which give the segments that shall be
        processed. Default is None, i.e., the whole file gets processed.
    @param current_sign_correction
        A dictionary. Its keys are the
        strings used in the file to indicate a state. The column from
        which this state is retrieved is given by #correction_column.
        The dictionaries' values are used to correct/normalize the
        current value in the file. For example, if discharge currents
        have the same positive sign as charge currents in the file, use
        -1 to correct that, or if the values are to be scaled by weight,
        use the scaling factor. The default is the empty dictionary.
    @param correction_column
        See #current_sign_correction. Default: -1.
    @param max_number_of_lines
        The maximum number of dataset lines that are to be read in.
        Default: -1 (no limit).
    @param flip_voltage_sign
        Defaults to False, where measured voltage remains unaltered.
        Change to True if the voltage shall be multiplied by -1.
        Also applies to impedances; real and imaginary parts.
    @param flip_imaginary_impedance_sign
        Defaults to False, where measured impedance remains unaltered.
        Change to True if the imaginary part of the impedance shall be
        multiplied by -1. Cancels out with 'flip-voltage-sign'.
    @return
        The measurement, packaged in a Measurement subclass. It depends
        on "datatype" which one it is:
         - "cycling": Cycling_Information
         - "static": Static_Information
         - "impedance": Impedance_Measurement
    """

    file = open(path, encoding=encoding)
    voltage_sign = -1.0 if flip_voltage_sign else 1.0
    imaginary_impedance_sign = -1.0 if flip_imaginary_impedance_sign else 1.0

    # Truncate the commented lines at the start.
    length_of_comments = 0
    for i in range(number_of_comment_lines):
        line = file.readline()
        length_of_comments = length_of_comments + len(line) + 1
    file.seek(length_of_comments)

    csv_file = csv.reader(file, delimiter=delimiter)
    data = [] if segment_column != -1 else [
        {index: [] for index in headers.keys()}
    ]
    last_segment_id = ""
    for i, row in enumerate(csv_file):
        if max_number_of_lines != -1 and i > max_number_of_lines:
            break
        try:
            if segment_column != -1 and row[segment_column] != last_segment_id:
                data.append({index: [] for index in headers.keys()})
                data[-1][correction_column] = []
                last_segment_id = row[segment_column]
        except IndexError:
            raise IndexError((
                "correction_column or segment_column are set to values outside"
                + " of the row range. Affected row: #{0}. Length of row: {1}. "
                + "Values for correction_column and segment_column: {2}, {3}. "
            ).format(i, len(row), correction_column, segment_column)
                + "Contents of affected row:" + linesep
                + linesep.join(
                    [str(j) + ' ' + str(entry) for j, entry in enumerate(row)]
            )
            )
        try:
            if correction_column != -1:
                data[-1][correction_column].append(
                    current_sign_correction[row[correction_column]]
                )
        except IndexError:
            raise IndexError((
                "correction_column is set to a value outside "
                + "of the row range. Affected row: #{0}. Length of row: {1}. "
                + "Value for correction_column: {2}. "
            ).format(i, len(row), correction_column)
                + "Contents of affected row:" + linesep
                + linesep.join(
                    [str(j) + ' ' + str(entry) for j, entry in enumerate(row)]
            )
            )
        except ValueError:
            raise ValueError((
                "current_sign_correction is missing an entry for the content "
                "of row #{0} at correction_column {1}. Content: {2}."
            ).format(i, correction_column, row[correction_column]))
        for index in headers.keys():
            try:
                data[-1][index].append(float(
                    row[index].replace(decimal, '.')
                ))
            except ValueError:  # Entry is not a number.
                data[-1][index].append(row[index])
            except IndexError:
                raise IndexError(
                    "Index " + str(index) + " out of bounds for row "
                    + str(row) + "."
                )

    file.close()

    header_info = {}
    for column_index, header in headers.items():
        # The following regex extracts the identifier, the unit prefix
        # and the unit. The "filter" removes the empty strings returned
        # from the regex, which are there for reversibility reasons.
        info = list(filter(None, re.split(
            r"([UItfϕ]?)(real_Z)?(imag_Z)?\s"
            r"\[([Mkmµn]??)([ΩVAhms]?)(Hz)?(rad)?(deg)?\]",
            header
        )))
        if len(info) == 2:
            id, unit = info
            prefix = ""
        elif len(info) == 3:
            id, prefix, unit = info
        else:
            id = header
            prefix = ""
            unit = ""
        header_info[id] = (column_index, prefix, unit)

    prefixes = {"M": 1e6, "k": 1e3, "": 1.0, "m": 1e-3, "µ": 1e-6, "n": 1e-9}

    if datatype in ["cycling", "static"]:
        try:
            timescale = {"h": 3600.0, "m": 60.0, "s": 1.0}[header_info["t"][2]]
        except KeyError:
            timescale = 1

        timepoints = []
        # timesteps = []
        currents = []
        voltages = []
        for index, segment in enumerate(data):
            if segments_to_process is not None:
                if index not in segments_to_process:
                    continue
            try:
                timepoints.append([
                    timescale * prefixes[header_info["t"][1]] * t
                    for t in segment[header_info["t"][0]]
                ])
            except KeyError:
                pass
            try:
                if correction_column != -1:
                    currents.append([
                        prefixes[header_info["I"][1]] * sign * current
                        for (current, sign) in zip(
                            segment[header_info["I"][0]],
                            segment[correction_column]
                        )
                    ])
                else:
                    currents.append([
                        prefixes[header_info["I"][1]] * current
                        for (current) in segment[header_info["I"][0]]
                    ])
            except KeyError:
                pass
            try:
                voltages.append([
                    prefixes[header_info["U"][1]] * voltage_sign * voltage
                    for voltage in segment[header_info["U"][0]]
                ])
            except KeyError:
                pass

        for h in ('t', 'I', 'U'):
            try:
                del header_info[h]
            except KeyError:
                pass
        other_columns = {}
        for id, column_index in header_info.items():
            other_columns[id] = []
            for segment in data:
                other_columns[id].append(segment[column_index[0]])

        if datatype == "cycling":
            return Cycling_Information(
                timepoints, currents, voltages, other_columns
            )
        elif datatype == "static":
            return Static_Information(
                timepoints, currents, voltages, other_columns
            )
    elif datatype == "impedance":
        frequencies = []
        real_impedances = []
        imaginary_impedances = []
        phases = []
        for index, segment in enumerate(data):
            if segments_to_process is not None:
                if index not in segments_to_process:
                    continue
            try:
                frequencies.append([
                    prefixes[header_info["f"][1]] * f
                    for f in segment[header_info["f"][0]]
                ])
            except KeyError:
                pass
            try:
                real_impedances.append([
                    prefixes[header_info["real_Z"][1]] * voltage_sign * imp
                    for imp in segment[header_info["real_Z"][0]]
                ])
            except KeyError:
                pass
            try:
                imaginary_impedances.append([
                    prefixes[header_info["imag_Z"][1]]
                    * voltage_sign
                    * imaginary_impedance_sign
                    * imp
                    for imp in segment[header_info["imag_Z"][0]]
                ])
            except KeyError:
                pass
            try:
                if header_info["ϕ"][2] == "deg":
                    phase_prefix = 2 * pi / 360
                elif header_info["ϕ"][2] == "rad":
                    phase_prefix = 1
                else:
                    raise ValueError(
                        "Phase angles have to be given in 'deg' or 'rad'."
                    )
                phases.append([
                    prefixes[header_info["ϕ"][1]] * phase_prefix * phase
                    for phase in segment[header_info["ϕ"][0]]
                ])
            except KeyError:
                pass

        for h in ('f', 'real_Z', 'imag_Z', 'ϕ'):
            try:
                del header_info[h]
            except KeyError:
                pass
        other_columns = {}
        for id, column_index in header_info.items():
            other_columns[id] = []
            for segment in data:
                other_columns[id].append(segment[column_index[0]])

        return Impedance_Measurement(
            frequencies,
            real_impedances,
            imaginary_impedances,
            phases,
            other_columns
        )
    else:
        raise ValueError(
            "Unsupported type of measurement: "
            + str(datatype)
            + ". Must be one of 'cycling', 'static', or 'impedance'."
        )


def print_hdf5_structure(
    h5py_object,
    depth=1,
    table_limit=4,
    verbose_limit=64
):
    """!brief Simple HDF5 structure viewer.
    @param h5py_object
        The HDF5 object wrapped by H5Py, for example h5py.File(filename, 'r').
    @param depth
        For pretty printing the recursive depth. Do not change.
    @param table_limit
        h5py.Dataset object tables will be truncated prior to printing up to
        this number in the higher dimension.
    @param verbose_limit
        h5py.Dataset objects will be truncated to at most this number in any
        dimension.
    """

    tree_visualizer = (depth - 1) * "  " + "|_"
    if isinstance(h5py_object, h5py.Dataset):
        print(tree_visualizer, h5py_object.shape, h5py_object.dtype)
        if len(h5py_object.shape) < 2:
            print(tree_visualizer, list(h5py_object[:verbose_limit]))
        else:
            if h5py_object.shape[0] < h5py_object.shape[1]:
                for column in h5py_object:
                    print(tree_visualizer, list(column[:table_limit]))
            else:
                for line in h5py_object[:table_limit]:
                    print(tree_visualizer, line)
        return
    elif (
        isinstance(h5py_object, h5py.Group)
        or isinstance(h5py_object, h5py.File)
    ):
        for key in h5py_object.keys():
            print(tree_visualizer[:-1])
            print(tree_visualizer, key)
            print_hdf5_structure(h5py_object[key], depth + 1)


def convert_none_notation_to_slicing(h5py_object, index):
    """!@brief Access a slice of an HDF5 object by transferable notation.
    @param h5py_object
        The HDF5 object wrapped by H5Py, for example h5py.File(filename, 'r').
    @param index
        A 2-tuple or a 2-list. (None, x) denotes slicing [:, x] and (x, None)
        denotes slicing [x, :].
    """

    if len(index) != 2:
        raise ValueError(
            "Location in HDF5 structure can only contain indices "
            "and 2-tuples or 2-lists denoting a [x ,y]-type slice."
        )
    if index[0] is None:
        return h5py_object[:, index[1]]
    elif index[1] is None:
        return h5py_object[index[0], :]
    else:
        raise ValueError(
            "2-tuple or 2-list in HDF5 structure location has to "
            "contain one None entry to denote the dimension of "
            "the slice that gets read completely."
        )


def get_hdf5_dataset_by_path(h5py_object, path):
    """!@brief Follow the structure of a HDF object to get a certain part.
    @param h5py_object
        The HDF5 object wrapped by H5Py, for example h5py.File(filename, 'r').
    @param path
        A list. Each entry goes one level deeper into the HDF structure. Each
        entry can either be the index to go into next itself, or a 2-tuple or a
        2-list. In the latter case, (None, x) denotes slicing [:, x] and
        (x, None) denotes slicing [x, :].
    @return
        Returns the HDF5 object found at the end of "path".
    """

    target_dataset = h5py_object
    for index in path:
        if isinstance(index, tuple) or isinstance(index, list):
            target_dataset = convert_none_notation_to_slicing(
                target_dataset, index
            )
        else:
            target_dataset = target_dataset[index]
    return target_dataset


def read_hdf5_table(
    path,
    data_location,
    headers,
    datatype="cycling",
    segment_location=None,
    segments_to_process=None,
    current_sign_correction={},
    correction_location=None,
    flip_voltage_sign=False,
    flip_imaginary_impedance_sign=False
):
    """!@brief Read the measurements as stored in a HDF5 file.
    @param path
        The full or relative path to the measurement file.
    @param data_location
        A list. Gives the location in the HDF5 file where the data table is
        stored. Set to None if everything is stored at the top level.
        Each entry goes one level deeper into the HDF structure. Each
        entry can either be the index to go into next itself, or a 2-tuple or a
        2-list. In the latter case, (None, x) denotes slicing [:, x] and
        (x, None) denotes slicing [x, :].
    @param headers
        A dictionary. Its keys are 2-tuples, slicing the data which are to
        be read in. Use the format "(x, None)" or "(None, x)" to slice
        the dimension with "None". The corresponding values are there to
        tell this function which kind of data is in which column.
        If necessary, the keys may also be tuples like 'data_location'.
        The following format has to be used: "<name> [<unit>]" where "name"
        is "U" (voltage), "I" (current) or "t" (time) and "unit" is
        "V", "A", "h", "m" or "s" with the optional prefixes "k", "m",
        "µ" or "n". This converts the data to prefix-less SI units.
        Additional columns may be read in with keys not in this format.
        The columns for segments and sign correction are only given by
        "segment_column" and "correction_column".
    @param datatype
        Default is "cycling", where cycling information is assumed in
        the file. "static" will trigger the additional extraction of
        exponential decays that are relevant to e.g. GITT.
        "impedance" will treat the file as an impedance measurement
        with frequencies and impedances instead of time and voltage.
    @segment_location
        A list, with the same format as "data_location". It points to the part
        of the data that stores the index of the current segment. If it changes
        from one data point to the next, that is used as the dividing line
        between two segments.
        Default is None, which returns the dataset in one segment.
    @param segments_to_process
        A list of indices which give the segments that shall be
        processed. Default is None, i.e., the whole file gets processed.
    @param current_sign_correction
        A dictionary. Its keys are the strings used in the file to indicate a
        state. The column from which this state is retrieved is given by
        "correction_column".
        The dictionaries' values are used to correct/normalize the
        current value in the file. For example, if discharge currents
        have the same positive sign as charge currents in the file, use
        -1 to correct that, or if the values are to be scaled by weight,
        use the scaling factor. The default is the empty dictionary.
    @param correction_location
        A list, with the same format as "data_location".
        For its use, see "current_sign_correction". Default: None.
    @param flip_voltage_sign
        Defaults to False, where measured voltage remains unaltered.
        Change to True if the voltage shall be multiplied by -1.
        Also applies to impedances; real and imaginary parts.
    @param flip_imaginary_impedance_sign
        Defaults to False, where measured impedance remains unaltered.
        Change to True if the imaginary part of the impedance shall be
        multiplied by -1. Cancels out with 'flip-voltage-sign'.
    @return
        The measurement, packaged in a Measurement subclass. It depends
        on "datatype" which one it is:
         - "cycling": Cycling_Information
         - "static": Static_Information
         - "impedance": Impedance_Measurement
    """

    file = h5py.File(path, 'r')
    voltage_sign = -1.0 if flip_voltage_sign else 1.0
    imaginary_impedance_sign = -1.0 if flip_imaginary_impedance_sign else 1.0
    data = [{}]

    header_info = {}
    for column_index, header in headers.items():
        # The following regex extracts the identifier, the unit prefix
        # and the unit. The "filter" removes the empty strings returned
        # from the regex, which are there for reversibility reasons.
        info = list(filter(None, re.split(
            r"([UItfϕ]?)(real_Z)?(imag_Z)?\s"
            r"\[([Mkmµn]??)([ΩVAhms]?)(Hz)?(rad)?(deg)?\]",
            header
        )))
        if len(info) == 2:
            id, unit = info
            prefix = ""
        elif len(info) == 3:
            id, prefix, unit = info
        else:
            id = header
            prefix = ""
            unit = ""
        header_info[id] = (column_index, prefix, unit)

    split_indices = [0]
    if segment_location:
        segment_data = get_hdf5_dataset_by_path(file, segment_location)
        last_segment_value = segment_data[0]
        for i, segment_value in enumerate(segment_data):
            if segment_value != last_segment_value:
                split_indices.append(i)
                data.append({})
                last_segment_value = segment_value
    split_indices.append(None)

    if data_location is None:
        data_table = file
    else:
        data_table = get_hdf5_dataset_by_path(file, data_location)
    for index in headers.keys():
        try:
            data_column = convert_none_notation_to_slicing(data_table, index)
        except ValueError:
            data_column = get_hdf5_dataset_by_path(data_table, index)
        for i, (start, stop) in enumerate(
            zip(split_indices[:-1], split_indices[1:])
        ):
            data[i][index] = data_column[start:stop]

    prefixes = {"M": 1e6, "k": 1e3, "": 1.0, "m": 1e-3, "µ": 1e-6, "n": 1e-9}

    if datatype in ["cycling", "static"]:
        try:
            timescale = {"h": 3600.0, "m": 60.0, "s": 1.0}[header_info["t"][2]]
        except KeyError:
            timescale = 1

        timepoints = []
        # timesteps = []
        currents = []
        voltages = []
        if correction_location:
            correction_data = get_hdf5_dataset_by_path(
                file, correction_location
            )
        for index, segment in enumerate(data):
            if segments_to_process is not None:
                if index not in segments_to_process:
                    continue
            try:
                timepoints.append([
                    timescale * prefixes[header_info["t"][1]] * t
                    for t in segment[header_info["t"][0]]
                ])
            except KeyError:
                pass
            # timesteps.append([t1 - t0
            #                   for (t0, t1) in zip(timepoints[-1][:-1],
            #                                       timepoints[-1][1:])])
            try:
                if correction_location:
                    for i, (start, stop) in enumerate(
                        zip(split_indices[:-1], split_indices[1:])
                    ):
                        currents.append([
                            prefixes[header_info["I"][1]]
                            * current_sign_correction[sign]
                            * current
                            for (current, sign) in zip(
                                segment[header_info["I"][0]],
                                correction_data[start:stop]
                            )
                        ])
                else:
                    currents.append([
                        prefixes[header_info["I"][1]] * current
                        for (current) in segment[header_info["I"][0]]
                    ])
            except KeyError:
                pass
            try:
                voltages.append([
                    prefixes[header_info["U"][1]] * voltage_sign * voltage
                    for voltage in segment[header_info["U"][0]]
                ])
            except KeyError:
                pass

        for h in ('t', 'I', 'U'):
            try:
                del header_info[h]
            except KeyError:
                pass
        other_columns = {}
        for id, column_index in header_info.items():
            other_columns[id] = []
            for segment in data:
                other_columns[id].append(segment[column_index[0]])

        if datatype == "cycling":
            return Cycling_Information(
                timepoints, currents, voltages, other_columns
            )
        elif datatype == "static":
            return Static_Information(
                timepoints, currents, voltages, other_columns
            )
    elif datatype == "impedance":
        frequencies = []
        real_impedances = []
        imaginary_impedances = []
        phases = []
        for index, segment in enumerate(data):
            if segments_to_process is not None:
                if index not in segments_to_process:
                    continue
            try:
                frequencies.append([
                    prefixes[header_info["f"][1]] * f
                    for f in segment[header_info["f"][0]]
                ])
            except KeyError:
                pass
            try:
                real_impedances.append([
                    prefixes[header_info["real_Z"][1]] * voltage_sign * imp
                    for imp in segment[header_info["real_Z"][0]]
                ])
            except KeyError:
                pass
            try:
                imaginary_impedances.append([
                    prefixes[header_info["imag_Z"][1]]
                    * voltage_sign
                    * imaginary_impedance_sign
                    * imp
                    for imp in segment[header_info["imag_Z"][0]]
                ])
            except KeyError:
                pass
            try:
                if header_info["ϕ"][2] == "deg":
                    phase_prefix = 2 * pi / 360
                elif header_info["ϕ"][2] == "rad":
                    phase_prefix = 1
                else:
                    raise ValueError(
                        "Phase angles have to be given in 'deg' or 'rad'."
                    )
                phases.append([
                    prefixes[header_info["ϕ"][1]] * phase_prefix * phase
                    for phase in segment[header_info["ϕ"][0]]
                ])
            except KeyError:
                pass

        for h in ('f', 'real_Z', 'imag_Z', 'ϕ'):
            try:
                del header_info[h]
            except KeyError:
                pass
        other_columns = {}
        for id, column_index in header_info.items():
            other_columns[id] = []
            for segment in data:
                other_columns[id].append(segment[column_index[0]])

        return Impedance_Measurement(
            frequencies,
            real_impedances,
            imaginary_impedances,
            phases,
            other_columns
        )
    else:
        raise ValueError(
            "Unsupported type of measurement: "
            + str(datatype)
            + ". Must be one of 'cycling', 'static', or 'impedance'."
        )



"""!@package ep_bolfi.tests.test_substitution_dict
Test suite for ep_bolfi.utility.preprocessing.SubstitutionDict.
"""


import unittest

from ep_bolfi.utility.preprocessing import SubstitutionDict


class TestSubstitutionDict(unittest.TestCase):
    """!@brief
    Test suite for ep_bolfi.utility.preprocessing.SubstitutionDict.
    """

    def initialize_test_dict(self):

        def func(passed_dict):
            return passed_dict['dolor'] + passed_dict['sit']

        base_dict = {
            "lorem": 1,
            "ipsum": 2.0,
        }
        substitutions = {
            "dolor": 'ipsum',
            "sit": lambda passed_dict: passed_dict['lorem'],
            "amet": func,
        }
        expected_result = {
            "lorem": 1,
            "ipsum": 2.0,
            "dolor": 2.0,
            "sit": 1,
            "amet": 3.0,
        }
        return (SubstitutionDict(base_dict, substitutions), expected_result)

    def test_getitem(self):
        test_dict, expected_result = self.initialize_test_dict()
        for k, v in test_dict.items():
            self.assertEqual(v, expected_result[k])

    def test_change(self):
        test_dict, _ = self.initialize_test_dict()
        test_dict['lorem'] = 2
        test_dict['ipsum'] = 4.0
        self.assertEqual(test_dict['dolor'], 4.0)
        self.assertEqual(test_dict['sit'], 2)
        self.assertEqual(test_dict['amet'], 6.0)


if __name__ == '__main__':
    unittest.main()
